import os
import warnings
from functools import lru_cache
from pathlib import Path
from tempfile import TemporaryDirectory
from time import gmtime, strftime
from typing import Iterator, List, NamedTuple, Optional, Tuple
from uuid import uuid4

import httpx

from ..dataset import Dataset, Layer, RemoteDataset
from ..utils import get_rich_progress
from ._resumable import Resumable
from .api_client.models import (
    ApiDatasetUploadInformation,
    ApiLinkedLayerIdentifier,
    ApiReserveDatasetUploadInformation,
)
from .context import _get_context, _WebknossosContext

DEFAULT_SIMULTANEOUS_UPLOADS = 5
MAXIMUM_RETRY_COUNT = 4


class LayerToLink(NamedTuple):
    dataset_name: str
    layer_name: str
    new_layer_name: Optional[str] = None
    organization_id: Optional[str] = (
        None  # defaults to the user's organization before uploading
    )

    @classmethod
    def from_remote_layer(
        cls,
        layer: Layer,
        new_layer_name: Optional[str] = None,
        organization_id: Optional[str] = None,
    ) -> "LayerToLink":
        ds = layer.dataset
        assert isinstance(
            ds, RemoteDataset
        ), f"The passed layer must belong to a RemoteDataset, but belongs to {ds}"
        return cls(ds._dataset_name, layer.name, new_layer_name, organization_id)

    def as_api_linked_layer_identifier(self) -> ApiLinkedLayerIdentifier:
        context = _get_context()
        return ApiLinkedLayerIdentifier(
            self.organization_id or context.organization_id,
            self.dataset_name,
            self.layer_name,
            self.new_layer_name,
        )


@lru_cache(maxsize=None)
def _cached_get_upload_datastore(context: _WebknossosContext) -> str:
    datastores = context.api_client_with_auth.datastore_list()
    for datastore in datastores:
        if datastore.allows_upload:
            return datastore.url
    raise ValueError("No datastore found where datasets can be uploaded.")


def _walk(
    path: Path,
    base_path: Optional[Path] = None,
) -> Iterator[Tuple[Path, Path, int]]:
    if base_path is None:
        base_path = path
    if path.is_dir():
        for p in path.iterdir():
            yield from _walk(p, base_path)
    else:
        yield (path.resolve(), path.relative_to(base_path), path.stat().st_size)


def upload_dataset(
    dataset: Dataset,
    new_dataset_name: Optional[str] = None,
    layers_to_link: Optional[List[LayerToLink]] = None,
    jobs: Optional[int] = None,
) -> str:
    if new_dataset_name is None:
        new_dataset_name = dataset.name
    if layers_to_link is None:
        layers_to_link = []
    context = _get_context()
    layer_names_to_link = set(i.new_layer_name or i.layer_name for i in layers_to_link)
    if len(layer_names_to_link.intersection(dataset.layers.keys())) > 0:
        warnings.warn(
            "[INFO] Excluding the following layers from upload, since they will be linked: "
            + f"{layer_names_to_link.intersection(dataset.layers.keys())}"
        )
        with TemporaryDirectory() as tmpdir:
            tmp_ds = dataset.shallow_copy_dataset(
                tmpdir, name=dataset.name, layers_to_ignore=layer_names_to_link
            )
            return upload_dataset(
                tmp_ds,
                new_dataset_name=new_dataset_name,
                layers_to_link=layers_to_link,
                jobs=jobs,
            )

    file_infos = list(_walk(dataset.path))
    total_file_size = sum(size for _, _, size in file_infos)
    # replicates https://github.com/scalableminds/webknossos/blob/master/frontend/javascripts/admin/dataset/dataset_upload_view.js
    time_str = strftime("%Y-%m-%dT%H-%M-%S", gmtime())
    upload_id = f"{time_str}__{uuid4()}"
    datastore_token = context.datastore_required_token
    datastore_url = _cached_get_upload_datastore(context)
    datastore_api_client = context.get_datastore_api_client(datastore_url)
    simultaneous_uploads = jobs if jobs is not None else DEFAULT_SIMULTANEOUS_UPLOADS
    if "PYTEST_CURRENT_TEST" in os.environ:
        simultaneous_uploads = 1
    is_valid_new_name_response = context.api_client_with_auth.dataset_is_valid_new_name(
        context.organization_id, new_dataset_name
    )
    if not is_valid_new_name_response.is_valid:
        problems_str = ""
        if is_valid_new_name_response.errors is not None:
            problems_str = f" Problems: {is_valid_new_name_response.errors}"
        raise Exception(
            f"Dataset name {context.organization_id}/{new_dataset_name} is not a valid new dataset name.{problems_str}"
        )

    datastore_api_client.dataset_reserve_upload(
        ApiReserveDatasetUploadInformation(
            upload_id,
            new_dataset_name,
            context.organization_id,
            total_file_count=len(file_infos),
            layers_to_link=[
                layer.as_api_linked_layer_identifier() for layer in layers_to_link
            ],
            folder_id=None,
            initial_teams=[],
        ),
        token=datastore_token,
        retry_count=MAXIMUM_RETRY_COUNT,
    )
    with get_rich_progress() as progress:
        with Resumable(
            f"{datastore_url}/data/datasets?token={datastore_token}",
            simultaneous_uploads=simultaneous_uploads,
            query={
                "owningOrganization": context.organization_id,
                "name": new_dataset_name,
                "totalFileCount": len(file_infos),
            },
            chunk_size=100 * 1024 * 1024,  # 100 MiB
            generate_unique_identifier=lambda _,
            relative_path: f"{upload_id}/{relative_path}",
            test_chunks=False,
            permanent_errors=[400, 403, 404, 409, 415, 500, 501],
            client=httpx.Client(timeout=None),
        ) as session:
            progress_task = progress.add_task("Dataset Upload", total=total_file_size)
            for file_path, relative_path, _ in file_infos:
                resumable_file = session.add_file(file_path, relative_path)
                resumable_file.chunk_completed.register(
                    lambda chunk: progress.advance(progress_task, chunk.size)
                )

    datastore_api_client.dataset_finish_upload(
        ApiDatasetUploadInformation(upload_id),
        datastore_token,
        retry_count=MAXIMUM_RETRY_COUNT,
    )

    return new_dataset_name



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple, Type, TypeVar, Union

import httpx

from ._serialization import custom_converter
from .errors import CannotHandleResponseError, UnexpectedStatusError

logger = logging.getLogger(__name__)

T = TypeVar("T")

Query = Dict[str, Optional[Union[str, int, float, bool]]]

LONG_TIMEOUT_SECONDS = 7200.0  # 2 hours


class AbstractApiClient(ABC):
    def __init__(
        self, timeout_seconds: float, headers: Optional[Dict[str, str]] = None
    ):
        self.headers = headers
        self.timeout_seconds = timeout_seconds

    @property
    @abstractmethod
    def url_prefix(self) -> str: ...

    def url_from_route(self, route: str) -> str:
        return f"{self.url_prefix}{route}"

    def _get_json(
        self, route: str, response_type: Type[T], query: Optional[Query] = None
    ) -> T:
        response = self._get(route, query)
        return self._parse_json(response, response_type)

    def _get_json_paginated(
        self,
        route: str,
        response_type: Type[T],
        limit: Optional[int],
        page_number: Optional[int],
        query: Optional[Query] = None,
    ) -> Tuple[T, int]:
        pagination_query: Query = {
            "limit": limit,
            "pageNumber": page_number,
            "includeTotalCount": True,
        }
        if query is not None:
            pagination_query.update(query)
        response = self._get(route, pagination_query)
        return self._parse_json(
            response, response_type
        ), self._extract_total_count_header(response)

    def _patch_json(self, route: str, body_structured: Any) -> None:
        body_json = self._prepare_for_json(body_structured)
        self._patch(route, body_json)

    def _post_json(
        self,
        route: str,
        body_structured: Any,
        query: Optional[Query] = None,
        retry_count: int = 0,
        timeout_seconds: Optional[float] = None,
    ) -> None:
        body_json = self._prepare_for_json(body_structured)
        self._post(
            route,
            body_json=body_json,
            query=query,
            retry_count=retry_count,
            timeout_seconds=timeout_seconds,
        )

    def _get_file(self, route: str, query: Optional[Query] = None) -> Tuple[bytes, str]:
        response = self._get(route, query)
        return response.content, self._parse_filename_from_header(response)

    def _post_with_json_response(self, route: str, response_type: Type[T]) -> T:
        response = self._post(route)
        return self._parse_json(response, response_type)

    def _post_json_with_json_response(
        self, route: str, body_structured: Any, response_type: Type[T]
    ) -> T:
        body_json = self._prepare_for_json(body_structured)
        response = self._post(route, body_json=body_json)
        return self._parse_json(response, response_type)

    def post_multipart_with_json_response(
        self,
        route: str,
        response_type: Type[T],
        multipart_data: Optional[httpx._types.RequestData] = None,
        files: Optional[httpx._types.RequestFiles] = None,
    ) -> T:
        response = self._post(route, multipart_data=multipart_data, files=files)
        return self._parse_json(response, response_type)

    def _get(
        self,
        route: str,
        query: Optional[Query] = None,
        timeout_seconds: Optional[float] = None,
    ) -> httpx.Response:
        return self._request("GET", route, query, timeout_seconds=timeout_seconds)

    def _patch(
        self,
        route: str,
        body_json: Optional[Any],
        query: Optional[Query] = None,
        timeout_seconds: Optional[float] = None,
    ) -> httpx.Response:
        return self._request(
            "PATCH",
            route,
            body_json=body_json,
            query=query,
            timeout_seconds=timeout_seconds,
        )

    def _post(
        self,
        route: str,
        body_json: Optional[Any] = None,
        query: Optional[Query] = None,
        multipart_data: Optional[httpx._types.RequestData] = None,
        files: Optional[httpx._types.RequestFiles] = None,
        retry_count: int = 0,
        timeout_seconds: Optional[float] = None,
    ) -> httpx.Response:
        return self._request(
            "POST",
            route,
            body_json=body_json,
            multipart_data=multipart_data,
            files=files,
            query=query,
            retry_count=retry_count,
            timeout_seconds=timeout_seconds,
        )

    def _request(
        self,
        method: str,
        route: str,
        query: Optional[Query] = None,
        body_json: Optional[Any] = None,
        multipart_data: Optional[httpx._types.RequestData] = None,
        files: Optional[httpx._types.RequestFiles] = None,
        retry_count: int = 0,
        timeout_seconds: Optional[float] = None,
    ) -> httpx.Response:
        assert (
            retry_count >= 0
        ), f"Cannot perform request with retry_count < 0, got {retry_count}"
        url = self.url_from_route(route)
        response = None
        for _ in range(retry_count + 1):
            response = httpx.request(
                method,
                url,
                params=self._omit_none_values_in_query(query),
                json=body_json,
                data=multipart_data,
                files=files,
                headers=self.headers,
                timeout=timeout_seconds or self.timeout_seconds,
            )
            if response.status_code == 200 or response.status_code == 400:
                # Stop retrying in case of success or bad request
                break
        assert response is not None, "Got no http response object"
        self._assert_good_response(response)
        return response

    def _omit_none_values_in_query(self, query: Optional[Query]) -> Optional[Query]:
        if query is None:
            return None
        return {k: v for (k, v) in query.items() if v is not None}

    def _parse_json(self, response: httpx.Response, response_type: Type[T]) -> T:
        try:
            return custom_converter.structure(response.json(), response_type)
        except Exception as e:
            raise CannotHandleResponseError(response) from e

    def _extract_total_count_header(self, response: httpx.Response) -> int:
        total_count_str = response.headers.get("X-Total-Count")
        assert total_count_str is not None, "X-Total-Count header missing from response"
        return int(total_count_str)

    def _parse_filename_from_header(self, response: httpx.Response) -> str:
        # Adapted from https://peps.python.org/pep-0594/#cgi
        from email.message import Message

        content_disposition_str = response.headers.get("content-disposition", "")

        m = Message()
        m["content-type"] = content_disposition_str
        return dict(m.get_params() or []).get("filename", "")

    def _prepare_for_json(self, body_structured: Any) -> Any:
        return custom_converter.unstructure(body_structured)

    def _assert_good_response(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise UnexpectedStatusError(response) from e



import mimetypes
import time
from pathlib import Path
from typing import Any, Dict

import httpx

from .file import FileChunk, ResumableFile
from .util import Config

RETRY_DELAY = 5  # seconds


class ResumableError(Exception):
    pass


def resolve_chunk(
    client: httpx.Client, config: Config, file: ResumableFile, chunk: FileChunk
) -> None:
    """Make sure a chunk is uploaded to the server and mark it as completed.

    Parameters
    ----------
    client : httpx.Client
        A client to use for communication with the server
    config : resumable.util.Config
        The configuration of the resumable client
    file : resumable.file.ResumableFile
        The parent file of the chunk to be resolved
    chunk : resumable.file.FileChunk
        The chunk to be resolved
    """

    exists_on_server = False
    if config.test_chunks:
        exists_on_server = _test_chunk(client, config, file, chunk)

    if not exists_on_server:
        tries = 0
        while not _send_chunk(client, config, file, chunk):
            time.sleep(RETRY_DELAY)
            tries += 1
            if tries >= config.max_chunk_retries:
                raise ResumableError("max retries exceeded")

    file.mark_chunk_completed(chunk)


def _test_chunk(
    client: httpx.Client, config: Config, file: ResumableFile, chunk: FileChunk
) -> bool:
    """Check if the chunk exists on the server.

    Returns
    -------
    bool
        True if the chunk exists on the server
    """
    response = client.get(
        config.target, params=_build_query(file, chunk, config.additional_query_params)
    )
    return response.status_code == 200


def _send_chunk(
    client: httpx.Client, config: Config, file: ResumableFile, chunk: FileChunk
) -> bool:
    """Upload the chunk to the server.

    Returns
    -------
    bool
        True if the upload was successful

    Raises
    ------
    ResumableError
        If the server responded with an error code indicating permanent failure
    """
    try:
        response = client.post(
            config.target,
            data=_build_query(file, chunk, config.additional_query_params),
            files={"file": chunk.read()},
        )
    except httpx.TransportError:
        # The request itself failed. The calling function will retry.
        return False
    if response.status_code in config.permanent_errors:
        raise ResumableError(f"{response.status_code} Error: {response.text}")
    return response.status_code in [200, 201]


def _build_query(
    file: ResumableFile, chunk: FileChunk, additional_query_params: Dict[str, str]
) -> Dict[str, Any]:
    """Build the query parameters for a chunk test or upload."""
    return {
        "resumableChunkSize": file.chunk_size,
        "resumableTotalSize": file.size,
        "resumableType": _file_type(file.path),
        "resumableIdentifier": str(file.unique_identifier),
        "resumableFilename": file.path.name,
        "resumableRelativePath": str(file.relative_path),
        "resumableTotalChunks": len(file.chunks),
        "resumableChunkNumber": chunk.index + 1,
        "resumableCurrentChunkSize": chunk.size,
        **additional_query_params,
    }


def _file_type(path: Path) -> str:
    """Mimic the type parameter of a JS File object.

    Resumable.js uses the File object's type attribute to guess mime type,
    which is guessed from file extention accoring to
    https://developer.mozilla.org/en-US/docs/Web/API/File/type.

    Parameters
    ----------
    path : str
        The path to guess the mime type of

    Returns
    -------
    str
        The inferred mime type, or '' if none could be inferred
    """
    type_, _ = mimetypes.guess_type(path.name)
    # When no type can be inferred, File.type returns an empty string
    return "" if type_ is None else type_



import json
import re
from typing import (
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

import attr
import numpy as np

from .mag import Mag
from .nd_bounding_box import NDBoundingBox
from .vec3_int import Vec3Int, Vec3IntLike

_DEFAULT_BBOX_NAME = "Unnamed Bounding Box"


@attr.frozen
class BoundingBox(NDBoundingBox):
    """
    This class is used to represent an axis-aligned cuboid in 3D.
    The top-left coordinate is inclusive and the bottom-right coordinate is exclusive.

    A small usage example:

    ```python
    from webknossos import BoundingBox

    bbox_1 = BoundingBox((0, 0, 0), (100, 100, 100))
    bbox_2 = BoundingBox((75, 75, 75), (100, 100, 100))

    assert bbox_1.intersected_with(bbox_2).size == (25, 25, 25)
    ```
    """

    topleft: Vec3Int = attr.field(converter=Vec3Int)
    size: Vec3Int = attr.field(converter=Vec3Int)
    axes: Tuple[str, str, str] = attr.field(default=("x", "y", "z"))
    index: Vec3Int = attr.field(default=Vec3Int(1, 2, 3))
    bottomright: Vec3Int = attr.field(init=False)
    name: Optional[str] = _DEFAULT_BBOX_NAME
    is_visible: bool = True
    color: Optional[Tuple[float, float, float, float]] = None

    def __attrs_post_init__(self) -> None:
        if not self.size.is_positive():
            # Flip the size in negative dimensions, so that the topleft is smaller than bottomright.
            # E.g. BoundingBox((10, 10, 10), (-5, 5, 5)) -> BoundingBox((5, 10, 10), (5, 5, 5)).
            negative_size = self.size.pairmin(Vec3Int.zeros())
            new_topleft = self.topleft + negative_size
            new_size = self.size.pairmax(-self.size)
            object.__setattr__(self, "topleft", new_topleft)
            object.__setattr__(self, "size", new_size)

        # Compute bottomright to avoid that it's recomputed every time
        # it is needed.
        object.__setattr__(self, "bottomright", self.topleft + self.size)

    def with_bounds_x(
        self, new_topleft_x: Optional[int] = None, new_size_x: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.x optionally replaced and size.x optionally replaced."""

        return cast(BoundingBox, self.with_bounds("x", new_topleft_x, new_size_x))

    def with_bounds_y(
        self, new_topleft_y: Optional[int] = None, new_size_y: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.y optionally replaced and size.y optionally replaced."""

        return cast(BoundingBox, self.with_bounds("y", new_topleft_y, new_size_y))

    def with_bounds_z(
        self, new_topleft_z: Optional[int] = None, new_size_z: Optional[int] = None
    ) -> "BoundingBox":
        """Returns a copy of the bounding box with topleft.z optionally replaced and size.z optionally replaced."""

        return cast(BoundingBox, self.with_bounds("z", new_topleft_z, new_size_z))

    @classmethod
    def from_wkw_dict(cls, bbox: Dict) -> "BoundingBox":
        return cls(bbox["topLeft"], [bbox["width"], bbox["height"], bbox["depth"]])

    @classmethod
    def from_config_dict(cls, bbox: Dict) -> "BoundingBox":
        return cls(bbox["topleft"], bbox["size"])

    @classmethod
    def from_tuple6(cls, tuple6: Tuple[int, int, int, int, int, int]) -> "BoundingBox":
        return cls(tuple6[0:3], tuple6[3:6])

    @classmethod
    def from_tuple2(cls, tuple2: Tuple[Vec3IntLike, Vec3IntLike]) -> "BoundingBox":
        return cls(tuple2[0], tuple2[1])

    @classmethod
    def from_points(cls, points: Iterable[Vec3IntLike]) -> "BoundingBox":
        """Returns a bounding box exactly containing all points."""

        all_points = np.array([Vec3Int(point).to_list() for point in points])
        topleft = all_points.min(axis=0)
        bottomright = all_points.max(axis=0)

        # bottomright is exclusive
        bottomright += 1

        return cls(topleft, bottomright - topleft)

    @classmethod
    def from_checkpoint_name(cls, checkpoint_name: str) -> "BoundingBox":
        """This function extracts a bounding box in the format `x_y_z_sx_sy_xz` which is contained in a string."""
        regex = r"(([0-9]+_){5}([0-9]+))"
        match = re.search(regex, checkpoint_name)
        assert (
            match is not None
        ), f"Could not extract bounding box from {checkpoint_name}"
        bbox_tuple = tuple(int(value) for value in match.group().split("_"))
        return cls.from_tuple6(cast(Tuple[int, int, int, int, int, int], bbox_tuple))

    @classmethod
    def from_csv(cls, csv_bbox: str) -> "BoundingBox":
        bbox_tuple = tuple(int(x) for x in csv_bbox.split(","))
        return cls.from_tuple6(cast(Tuple[int, int, int, int, int, int], bbox_tuple))

    @classmethod
    def from_ndbbox(cls, bbox: NDBoundingBox) -> "BoundingBox":
        return cls(bbox.topleft_xyz, bbox.size_xyz)

    @classmethod
    def from_auto(
        cls, obj: Union["BoundingBox", str, Dict, List, Tuple]
    ) -> "BoundingBox":
        if isinstance(obj, cls):
            return obj
        elif isinstance(obj, str):
            if ":" in obj:
                return cls.from_auto(json.loads(obj))
            else:
                return cls.from_csv(obj)
        elif isinstance(obj, dict):
            if "size" in obj:
                return cls.from_config_dict(obj)
            return cls.from_wkw_dict(obj)
        elif isinstance(obj, list) or isinstance(obj, tuple):
            if len(obj) == 2:
                return cls.from_tuple2(obj)  # type: ignore
            elif len(obj) == 6:
                return cls.from_tuple6(obj)  # type: ignore

        raise Exception("Unknown bounding box format.")

    @classmethod
    def empty(
        cls,
    ) -> "BoundingBox":
        return cls(Vec3Int.zeros(), Vec3Int.zeros())

    def to_wkw_dict(self) -> dict:
        (
            width,
            height,
            depth,
        ) = self.size.to_list()

        return {
            "topLeft": self.topleft.to_list(),
            "width": width,
            "height": height,
            "depth": depth,
        }

    def to_config_dict(self) -> dict:
        return {"topleft": self.topleft.to_list(), "size": self.size.to_list()}

    def to_checkpoint_name(self) -> str:
        x, y, z = self.topleft
        width, height, depth = self.size
        return "{x}_{y}_{z}_{width}_{height}_{depth}".format(
            x=x, y=y, z=z, width=width, height=height, depth=depth
        )

    def to_tuple6(self) -> Tuple[int, int, int, int, int, int]:
        return tuple(self.topleft.to_list() + self.size.to_list())  # type: ignore

    def to_csv(self) -> str:
        return ",".join(map(str, self.to_tuple6()))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, NDBoundingBox):
            self._check_compatibility(other)
            return self.topleft == other.topleft and self.size == other.size

        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"BoundingBox(topleft={self.topleft.to_tuple()}, size={self.size.to_tuple()})"

    def padded_with_margins(
        self, margins_left: Vec3IntLike, margins_right: Optional[Vec3IntLike] = None
    ) -> "BoundingBox":
        if margins_right is None:
            margins_right = margins_left

        margins_left = Vec3Int(margins_left)
        margins_right = Vec3Int(margins_right)

        return attr.evolve(
            self,
            topleft=self.topleft - margins_left,
            size=self.size + (margins_left + margins_right),
        )

    def is_empty(self) -> bool:
        return not self.size.is_positive(strictly_positive=True)

    def in_mag(self, mag: Mag) -> "BoundingBox":
        mag_vec = mag.to_vec3_int()

        assert (
            self.topleft % mag_vec == Vec3Int.zeros()
        ), f"topleft {self.topleft} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."
        assert (
            self.bottomright % mag_vec == Vec3Int.zeros()
        ), f"bottomright {self.bottomright} is not aligned with the mag {mag}. Use BoundingBox.align_with_mag()."

        return attr.evolve(
            self,
            topleft=(self.topleft // mag_vec),
            size=(self.size // mag_vec),
        )

    def _align_with_mag_slow(self, mag: Mag, ceil: bool = False) -> "BoundingBox":
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrinked when necessary.
        """
        np_mag = mag.to_np()

        def align(point: Vec3Int, round_fn: Callable) -> Vec3Int:
            return round_fn(point.to_np() / np_mag).astype(int) * np_mag

        if ceil:
            topleft = align(self.topleft, np.floor)
            bottomright = align(self.bottomright, np.ceil)
        else:
            topleft = align(self.topleft, np.ceil)
            bottomright = align(self.bottomright, np.floor)
        return attr.evolve(self, topleft=topleft, size=bottomright - topleft)

    def align_with_mag(
        self, mag: Union[Mag, Vec3Int], ceil: bool = False
    ) -> "BoundingBox":
        """Rounds the bounding box, so that both topleft and bottomright are divisible by mag.

        :argument ceil: If true, the bounding box is enlarged when necessary. If false, it's shrinked when necessary.
        """
        # This does the same as _align_with_mag_slow, which is more readable.
        # Same behavior is asserted in test_align_with_mag_against_numpy_implementation
        mag_vec = mag.to_vec3_int() if isinstance(mag, Mag) else mag
        roundup = self.topleft if ceil else self.bottomright
        rounddown = self.bottomright if ceil else self.topleft
        margin_to_roundup = roundup % mag_vec
        aligned_roundup = roundup - margin_to_roundup
        margin_to_rounddown = (mag_vec - (rounddown % mag_vec)) % mag_vec
        aligned_rounddown = rounddown + margin_to_rounddown
        if ceil:
            return attr.evolve(
                self, topleft=aligned_roundup, size=aligned_rounddown - aligned_roundup
            )
        else:
            return attr.evolve(
                self,
                topleft=aligned_rounddown,
                size=aligned_roundup - aligned_rounddown,
            )

    def contains(self, coord: Union[Vec3IntLike, np.ndarray]) -> bool:
        """Check whether a point is inside of the bounding box.
        Note that the point may have float coordinates in the ndarray case"""

        if isinstance(coord, np.ndarray):
            assert (
                coord.shape == (3,)
            ), f"Numpy array BoundingBox.contains must have shape (3,), got {coord.shape}."
            return cast(
                bool,
                np.all(coord >= self.topleft) and np.all(coord < self.bottomright),
            )
        else:
            # In earlier versions, we simply converted to ndarray to have
            # a unified calculation here, but this turned out to be a performance bottleneck.
            # Therefore, the contains-check is performed on the tuple here.
            coord = Vec3Int(coord)
            return (
                self.topleft[0] <= coord[0] < self.bottomright[0]
                and self.topleft[1] <= coord[1] < self.bottomright[1]
                and self.topleft[2] <= coord[2] < self.bottomright[2]
            )

    def chunk(
        self,
        chunk_shape: Vec3IntLike,
        chunk_border_alignments: Optional[Vec3IntLike] = None,
    ) -> Generator["BoundingBox", None, None]:
        """Decompose the bounding box into smaller chunks of size `chunk_shape`.

        Chunks at the border of the bounding box might be smaller than chunk_shape.
        If `chunk_border_alignment` is set, all border coordinates
        *between two chunks* will be divisible by that value.
        """

        start = self.topleft.to_np()
        chunk_shape = Vec3Int(chunk_shape).to_np()

        start_adjust = np.array([0, 0, 0])
        if chunk_border_alignments is not None:
            chunk_border_alignments_array = Vec3Int(chunk_border_alignments).to_np()
            assert np.all(
                chunk_shape % chunk_border_alignments_array == 0
            ), f"{chunk_shape} not divisible by {chunk_border_alignments_array}"

            # Move the start to be aligned correctly. This doesn't actually change
            # the start of the first chunk, because we'll intersect with `self`,
            # but it'll lead to all chunk borders being aligned correctly.
            start_adjust = start % chunk_border_alignments_array

        for x in range(
            start[0] - start_adjust[0], start[0] + self.size[0], chunk_shape[0]
        ):
            for y in range(
                start[1] - start_adjust[1], start[1] + self.size[1], chunk_shape[1]
            ):
                for z in range(
                    start[2] - start_adjust[2], start[2] + self.size[2], chunk_shape[2]
                ):
                    yield cast(
                        BoundingBox,
                        BoundingBox([x, y, z], chunk_shape).intersected_with(self),
                    )

    def offset(self, vector: Vec3IntLike) -> "BoundingBox":
        return attr.evolve(self, topleft=self.topleft + Vec3Int(vector))

    def __hash__(self) -> int:
        return hash(self.to_tuple6())



import warnings
from typing import TYPE_CHECKING, List, Optional

import attr

from ..client.api_client.models import ApiProject
from ..client.context import _get_api_client
from .user import User

if TYPE_CHECKING:
    from .task import Task


@attr.frozen
class Project:
    """Data class containing information about a WEBKNOSSOS project"""

    project_id: str
    name: str
    team_id: str
    team_name: str
    owner_id: Optional[str]  # None in case you have no read access on the owner
    priority: int
    paused: bool
    expected_time: Optional[int]

    @classmethod
    def get_by_id(cls, project_id: str) -> "Project":
        """Returns the project specified by the passed id if your token authorizes you to see it."""
        api_project = _get_api_client(enforce_auth=True).project_info_by_id(project_id)
        return cls._from_api_project(api_project)

    @classmethod
    def get_by_name(cls, name: str) -> "Project":
        """Returns the user specified by the passed name if your token authorizes you to see it."""
        api_project = _get_api_client(enforce_auth=True).project_info_by_name(name)
        return cls._from_api_project(api_project)

    def get_tasks(self, fetch_all: bool = False) -> List["Task"]:
        """Returns the tasks of this project.
        Note: will fetch only the first 1000 entries by default, warns if that means some are missing.
        set parameter pass fetch_all=True to use pagination to fetch all tasks iteratively with pagination.
        """

        from .task import Task

        PAGINATION_LIMIT = 1000
        pagination_page = 0

        client = _get_api_client(enforce_auth=True)
        api_tasks_batch, total_count = client.task_infos_by_project_id_paginated(
            self.project_id, limit=PAGINATION_LIMIT, page_number=pagination_page
        )
        all_tasks = [Task._from_api_task(t) for t in api_tasks_batch]
        if total_count > PAGINATION_LIMIT:
            if fetch_all:
                while total_count > len(all_tasks):
                    pagination_page += 1
                    api_tasks_batch, _ = client.task_infos_by_project_id_paginated(
                        self.project_id,
                        limit=PAGINATION_LIMIT,
                        page_number=pagination_page,
                    )
                    new_tasks = [Task._from_api_task(t) for t in api_tasks_batch]
                    all_tasks.extend(new_tasks)

            else:
                warnings.warn(
                    f"[INFO] Fetched only {PAGINATION_LIMIT} of {total_count} tasks. Pass fetch_all=True to fetch all tasks iteratively (may be slow!)"
                )

        return all_tasks

    def get_owner(self) -> User:
        """Returns the user that is the owner of this task"""
        assert (
            self.owner_id is not None
        ), "Project owner is None, you may not have enough access rights to read the project owner."
        return User.get_by_id(self.owner_id)

    @classmethod
    def _from_api_project(cls, api_project: ApiProject) -> "Project":
        owner_id = None
        if api_project.owner is not None:
            owner_id = api_project.owner.id
        return cls(
            api_project.id,
            api_project.name,
            api_project.team,
            api_project.team_name,
            owner_id,
            api_project.priority,
            api_project.paused,
            api_project.expected_time,
        )



from typing import NamedTuple
from xml.etree.ElementTree import Element

from loxun import XmlWriter

from .utils import enforce_not_null


class Meta(NamedTuple):
    name: str
    content: str

    def _dump(self, xf: XmlWriter) -> None:
        xf.tag(
            "meta",
            {"name": self.name, "content": self.content},
        )

    @classmethod
    def _parse(cls, nml_meta: Element) -> "Meta":
        return cls(
            enforce_not_null(nml_meta.get("name", default=None)),
            enforce_not_null(nml_meta.get("content", default=None)),
        )



"""Annotations can contain annotated data in two forms:
  - [skeleton data](/webknossos/skeleton_annotation.html), as provided by the `Skeleton` class, and
  - [volume annotation layers](/webknossos/volume_annotation.html) (or volume layers short),
    which can be exported as a `SegmentationLayer`, see `export_volume_layer_to_dataset()`
    and `temporary_volume_layer_copy()`.

Usually, annotations should be created manually in the WEBKNOSSOS interface and can be downloaded using
`Annotation.download()`. The downloaded instance is not persisted to disk automatically, please use `save()`
for this purpose. The general purpose file format is `.zip` files containing an `.nml` file with
meta-information and the skeleton data and also containing inner `.zip` files for the volume layers.
For skeleton-only annotations without volume layers `.nml` files can be used directly. Both formats
are compatible with the WEBKNOSSOS up- and downloads.

To prepare volume annotations in the code for correction of segmentation data in the WEBKNOSSOS interface,
please use `add_volume_layer()` with the `fallback_layer` argument, referencing a segmentation layer that
is available on WEBKNOSSOS (e.g. using the `Dataset` upload before).
Correcting segmentations using fallback layers is much more efficient, adding volume
annotation data programmatically is discouraged therefore.
"""

import json
import re
import warnings
from contextlib import contextmanager, nullcontext
from enum import Enum, unique
from io import BytesIO
from os import PathLike
from pathlib import Path
from shutil import copyfileobj
from tempfile import TemporaryDirectory
from typing import (
    BinaryIO,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
    overload,
)
from zipfile import ZIP_DEFLATED, ZipFile
from zlib import Z_BEST_SPEED

import attr
from upath import UPath
from zipp import Path as ZipPath

import webknossos._nml as wknml

from ..dataset import (
    SEGMENTATION_CATEGORY,
    DataFormat,
    Dataset,
    Layer,
    RemoteDataset,
    SegmentationLayer,
)
from ..dataset.defaults import PROPERTIES_FILE_NAME
from ..dataset.properties import DatasetProperties, dataset_converter
from ..geometry import NDBoundingBox, Vec3Int
from ..skeleton import Skeleton
from ..utils import time_since_epoch_in_ms, warn_deprecated
from ._nml_conversion import annotation_to_nml, nml_to_skeleton

Vector3 = Tuple[float, float, float]
Vector4 = Tuple[float, float, float, float]


MAG_RE = r"((\d+-\d+-)?\d+)"
SEP_RE = r"(\/|\\)"
CUBE_RE = rf"z\d+{SEP_RE}y\d+{SEP_RE}x\d+\.wkw"
ANNOTATION_WKW_PATH_RE = re.compile(rf"{MAG_RE}{SEP_RE}(header\.wkw|{CUBE_RE})")


@attr.define
class SegmentInformation:
    name: Optional[str]
    anchor_position: Optional[Vec3Int]
    color: Optional[Vector4]


@attr.define
class _VolumeLayer:
    id: int
    name: str
    fallback_layer_name: Optional[str]
    data_format: DataFormat
    zip: Optional[ZipPath]
    segments: Dict[int, SegmentInformation]
    largest_segment_id: Optional[int]

    def _default_zip_name(self) -> str:
        return f"data_{self.id}_{self.name}.zip"


def _extract_zip_folder(zip_file: ZipFile, out_path: Path, prefix: str) -> None:
    for zip_entry in zip_file.filelist:
        if zip_entry.filename.startswith(prefix) and not zip_entry.is_dir():
            out_file_path = out_path / (zip_entry.filename[len(prefix) :])
            out_file_path.parent.mkdir(parents=True, exist_ok=True)
            with zip_file.open(zip_entry, "r") as zip_f, out_file_path.open(
                "wb"
            ) as out_f:
                copyfileobj(zip_f, out_f)


@attr.define
class Annotation:
    name: str
    skeleton: Skeleton = None  # type: ignore[assignment]
    # The following underscored attributes are just for initialization
    # in case the skeleton is not given. They are always None as attributes.
    _dataset_name: Optional[str] = None
    _voxel_size: Optional[Vector3] = None
    _organization_id: Optional[str] = None
    _description: Optional[str] = None
    owner_name: Optional[str] = None
    annotation_id: Optional[str] = None
    time: Optional[int] = attr.ib(factory=time_since_epoch_in_ms)
    edit_position: Optional[Vector3] = None
    edit_rotation: Optional[Vector3] = None
    zoom_level: Optional[float] = None
    metadata: Dict[str, str] = attr.Factory(dict)
    task_bounding_box: Optional[NDBoundingBox] = None
    user_bounding_boxes: List[NDBoundingBox] = attr.Factory(list)
    _volume_layers: List[_VolumeLayer] = attr.field(factory=list, init=False)

    @classmethod
    def _set_init_docstring(cls) -> None:
        Annotation.__init__.__doc__ = """
        To initialize a local annotation, please provide the `name` argument, and either
        the `skeleton` argument, or a `dataset_name` and `voxel_size`.
        When supplying `skeleton` passing `dataset_name`, `voxel_size`, `organization_id` or
        `description` is not allowed as the attributes of the skeleton are used in this case.
        """

    def __attrs_post_init__(self) -> None:
        if self.skeleton is None:
            assert (
                self._dataset_name is not None
            ), "Please either supply a skeleton or dataset_name for Annotation()."
            assert (
                self._voxel_size is not None
            ), "Please supply a voxel_size for Annotation()."
            self.skeleton = Skeleton(
                dataset_name=self._dataset_name,
                voxel_size=self._voxel_size,
                organization_id=self._organization_id,
                description=self._description,
            )
            self._dataset_name = None
            self._voxel_size = None
            self._organization_id = None
            self._description = None
        else:
            assert all(
                i is None
                for i in [
                    self._dataset_name,
                    self._voxel_size,
                    self._organization_id,
                    self._description,
                ]
            ), (
                "When supplying a skeleton for Annotation(), passing dataset_name, voxel_size, organization_id or description is not allowed. "
                + "The attributes of the skeleton are used in this case."
            )

    @property
    def username(self) -> Optional[str]:
        """Deprecated, use `owner_name` instead."""
        warn_deprecated("username", "owner_name")
        return self.owner_name

    @username.setter
    def username(self, username: str) -> None:
        warn_deprecated("username", "owner_name")
        self.owner_name = username

    @property
    def dataset_name(self) -> str:
        """This attribute is a proxy for `skeleton.dataset_name`."""
        return self.skeleton.dataset_name

    @dataset_name.setter
    def dataset_name(self, dataset_name: str) -> None:
        self.skeleton.dataset_name = dataset_name

    @property
    def voxel_size(self) -> Tuple[float, float, float]:
        """This attribute is a proxy for `skeleton.voxel_size`."""
        return self.skeleton.voxel_size

    @voxel_size.setter
    def voxel_size(self, voxel_size: Tuple[float, float, float]) -> None:
        self.skeleton.voxel_size = voxel_size

    @property
    def scale(self) -> Tuple[float, float, float]:
        """Deprecated, please use `voxel_size`."""
        warn_deprecated("scale", "voxel_size")
        return self.voxel_size

    @scale.setter
    def scale(self, scale: Tuple[float, float, float]) -> None:
        """Deprecated, please use `voxel_size`."""
        warn_deprecated("scale", "voxel_size")
        self.voxel_size = scale

    @property
    def organization_id(self) -> Optional[str]:
        """This attribute is a proxy for `skeleton.organization_id`."""
        return self.skeleton.organization_id

    @organization_id.setter
    def organization_id(self, organization_id: Optional[str]) -> None:
        self.skeleton.organization_id = organization_id

    @property
    def description(self) -> Optional[str]:
        """This attribute is a proxy for `skeleton.description`."""
        return self.skeleton.description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self.skeleton.description = description

    @classmethod
    def load(cls, annotation_path: Union[str, PathLike]) -> "Annotation":
        """Loads a `.nml` file or a `.zip` file containing an NML and possibly also volume
        layers. Returns the `Annotation` object."""
        annotation_path = Path(annotation_path)
        assert (
            annotation_path.exists()
        ), f"Annotation path {annotation_path} does not exist."
        if annotation_path.suffix == ".zip":
            return cls._load_from_zip(annotation_path)
        elif annotation_path.suffix == ".nml":
            with annotation_path.open(mode="rb") as f:
                return cls._load_from_nml(annotation_path.stem, f)
        else:
            raise RuntimeError(
                "The loaded annotation must have the suffix .zip or .nml, but is {annotation_path.suffix}"
            )

    @overload
    @classmethod
    def download(
        cls,
        annotation_id_or_url: str,
        annotation_type: Union[str, "AnnotationType", None] = None,
        webknossos_url: Optional[str] = None,
        *,
        skip_volume_data: bool = False,
    ) -> "Annotation": ...

    @overload
    @classmethod
    def download(
        cls,
        annotation_id_or_url: str,
        annotation_type: Union[str, "AnnotationType", None] = None,
        webknossos_url: Optional[str] = None,
        *,
        skip_volume_data: bool = False,
        _return_context: bool,
    ) -> Tuple["Annotation", ContextManager[None]]: ...

    @classmethod
    def download(
        cls,
        annotation_id_or_url: str,
        annotation_type: Union[str, "AnnotationType", None] = None,
        webknossos_url: Optional[str] = None,
        *,
        skip_volume_data: bool = False,
        _return_context: bool = False,
    ) -> Union["Annotation", Tuple["Annotation", ContextManager[None]]]:
        """
        * `annotation_id_or_url` may be an annotation id or a full URL to an annotation, e.g.
          `https://webknossos.org/annotations/6114d9410100009f0096c640`
        * `annotation_type` is no longer required and therefore deprecated and ignored
        * `webknossos_url` may be supplied if an annotation id was used
          and allows to specifiy in which webknossos instance to search for the annotation.
          It defaults to the url from your current `webknossos_context`, using https://webknossos.org as a fallback.
        * `skip_volume_data` can be set to `True` to omit downloading annotated volume data.
          They can still be streamed from WEBKNOSSOS using `annotation.get_remote_annotation_dataset()`.
        * `_return_context` should not be set.
        """
        from ..client._resolve_short_link import resolve_short_link
        from ..client.context import _get_api_client, _get_context, webknossos_context

        annotation_id_or_url = resolve_short_link(annotation_id_or_url)

        match = re.match(_ANNOTATION_URL_REGEX, annotation_id_or_url)
        if match is not None:
            assert webknossos_url is None and annotation_type is None, (
                "When Annotation.download() is be called with an annotation url, "
                + "e.g. Annotation.download('https://webknossos.org/annotations/6114d9410100009f0096c640'), "
                + "annotation_type and webknossos_url must not be set."
            )
            annotation_id = match.group("annotation_id")
            webknossos_url = match.group("webknossos_url")
        else:
            annotation_id = annotation_id_or_url

        if annotation_type is not None:
            warnings.warn(
                "[DEPRECATION] `annotation_type` is deprecated for Annotation.download(), it should be omitted.",
                DeprecationWarning,
            )

        if webknossos_url is not None and webknossos_url != _get_context().url:
            warnings.warn(
                f"[INFO] The supplied url {webknossos_url} does not match your current context {_get_context().url}. "
                + "Using no token, only public annotations can be downloaded. "
                + "Please see https://docs.webknossos.org/api/webknossos/client/context.html to adapt the URL and token."
            )
            context: ContextManager[None] = webknossos_context(
                webknossos_url, token=None
            )
        else:
            context = nullcontext()

        with context:
            client = _get_api_client()
            file_body, filename = client.annotation_download(
                annotation_id, skip_volume_data=skip_volume_data
            )

        if filename.endswith(".nml"):
            annotation = Annotation._load_from_nml(filename[:-4], BytesIO(file_body))
        else:
            assert filename.endswith(
                ".zip"
            ), f"Downloaded annoation should have the suffix .zip or .nml, but has filename {filename}"
            annotation = Annotation._load_from_zip(BytesIO(file_body))

        if _return_context:
            return annotation, context
        else:
            return annotation

    @classmethod
    def open_as_remote_dataset(
        cls,
        annotation_id_or_url: str,
        annotation_type: Union[str, "AnnotationType", None] = None,
        webknossos_url: Optional[str] = None,
    ) -> Dataset:
        (
            annotation,
            context,
        ) = Annotation.download(
            annotation_id_or_url=annotation_id_or_url,
            annotation_type=annotation_type,
            webknossos_url=webknossos_url,
            skip_volume_data=True,
            _return_context=True,
        )

        with context:
            return annotation.get_remote_annotation_dataset()

    @classmethod
    def _load_from_nml(
        cls,
        name: str,
        nml_content: BinaryIO,
        possible_volume_paths: Optional[List[ZipPath]] = None,
    ) -> "Annotation":
        nml = wknml.Nml.parse(nml_content)

        annotation = cls(
            name=name,
            skeleton=nml_to_skeleton(nml),
            owner_name=nml.get_meta("username"),
            annotation_id=nml.get_meta("annotationId"),
            time=nml.parameters.time,
            edit_position=nml.parameters.editPosition,
            edit_rotation=nml.parameters.editRotation,
            zoom_level=nml.parameters.zoomLevel,
            task_bounding_box=nml.parameters.taskBoundingBox,
            user_bounding_boxes=nml.parameters.userBoundingBoxes or [],
            metadata={
                i.name: i.content
                for i in nml.meta
                if i.name not in ["username", "annotationId"]
            },
        )
        annotation._volume_layers = cls._parse_volumes(nml, possible_volume_paths)
        return annotation

    @staticmethod
    def _parse_volumes(
        nml: wknml.Nml, possible_paths: Optional[List[ZipPath]]
    ) -> List[_VolumeLayer]:
        volume_layers = []
        layers_with_not_found_location = []
        layers_without_location = []
        for volume in nml.volumes:
            if possible_paths is None:  # when parsing NML files
                volume_path = None
                if volume.location is not None:
                    # This should only happen if a zipped nml
                    # is unpacked and loaded directly.
                    layers_with_not_found_location.append(volume)
                    volume_path = None
            elif volume.location is None:
                volume_path = None
                layers_without_location.append(volume)
            else:
                fitting_volume_paths = [
                    i for i in possible_paths if str(i.at) == volume.location
                ]
                if len(fitting_volume_paths) == 1:
                    with fitting_volume_paths[0].open(mode="rb") as f:
                        with ZipFile(f) as volume_layer_zipfile:
                            if len(volume_layer_zipfile.filelist) == 0:
                                volume_path = None
                            else:
                                volume_path = fitting_volume_paths[0]
                else:
                    layers_with_not_found_location.append(volume)
                    volume_path = None

            segments = {}
            if volume.segments is not None:
                for segment in volume.segments:
                    segments[segment.id] = SegmentInformation(
                        name=segment.name,
                        anchor_position=segment.anchor_position,
                        color=segment.color,
                    )
            volume_layers.append(
                _VolumeLayer(
                    id=volume.id,
                    name="Volume" if volume.name is None else volume.name,
                    fallback_layer_name=volume.fallback_layer,
                    data_format=DataFormat(volume.format)
                    if volume.format is not None
                    else DataFormat.WKW,
                    zip=volume_path,
                    segments=segments,
                    largest_segment_id=volume.largest_segment_id,
                )
            )
        assert len(set(i.id for i in volume_layers)) == len(
            volume_layers
        ), "Some volume layers have the same id, this is not allowed."
        if len(layers_without_location) > 0:
            warnings.warn(
                "[INFO] Omitting the volume layer annotation data for layers "
                + f"{[v.name or v.id for v in layers_without_location]}, "
                + "as their location is not referenced in the NML."
            )
        if len(layers_with_not_found_location) > 0:
            warnings.warn(
                "[INFO] Omitting the volume layer annotation data for layers "
                + f"{[v.name or v.id for v in layers_without_location]}, "
                + f"as their referenced files {[v.location for v in layers_without_location]} "
                + "cannot be found."
            )
        return volume_layers

    @classmethod
    def _load_from_zip(cls, content: Union[str, PathLike, BinaryIO]) -> "Annotation":
        zipfile = ZipFile(content)
        paths = [ZipPath(zipfile, i.filename) for i in zipfile.filelist]
        nml_paths = [i for i in paths if i.suffix == ".nml"]
        assert len(nml_paths) > 0, "Couldn't find an nml file in the supplied zip-file."
        assert (
            len(nml_paths) == 1
        ), f"There must be exactly one nml file in the zip-file, but found {len(nml_paths)}."
        with nml_paths[0].open(mode="rb") as f:
            return cls._load_from_nml(nml_paths[0].stem, f, possible_volume_paths=paths)

    def save(self, path: Union[str, PathLike]) -> None:
        """
        Stores the annotation as a zip or nml at the given path.
        """
        path = Path(path)
        assert path.suffix in [
            ".zip",
            ".nml",
        ], "The target path must have a .zip or .nml suffix."

        if path.suffix == ".zip":
            with ZipFile(
                path, mode="x", compression=ZIP_DEFLATED, compresslevel=Z_BEST_SPEED
            ) as zipfile:
                self._write_to_zip(zipfile)
        else:
            assert len(self._volume_layers) == 0, (
                f"Annotation {self.name} contains volume annotations and cannot be saved as an NML file. "
                + "Please use a .zip path instead."
            )
            nml = annotation_to_nml(self)
            with open(path, "wb") as f:
                nml.write(f)

    def upload(self) -> str:
        """Uploads the annotation to your current `webknossos_context`."""
        from ..client.context import _get_api_client, _get_context

        context = _get_context()

        client = _get_api_client(enforce_auth=True)
        response_annotation_info = client.annotation_upload(
            self._binary_zip(), f"{self.name}.zip", createGroupForEachFile=False
        )

        return f"{context.url}/annotations/{response_annotation_info.annotation.typ}/{response_annotation_info.annotation.id}"

    def _binary_zip(self) -> bytes:
        with BytesIO() as buffer:
            with ZipFile(
                buffer, mode="a", compression=ZIP_DEFLATED, compresslevel=Z_BEST_SPEED
            ) as zipfile:
                self._write_to_zip(zipfile)
            return buffer.getvalue()

    def _write_to_zip(self, zipfile: ZipFile) -> None:
        nml = annotation_to_nml(self)
        with BytesIO() as buffer:
            nml.write(buffer)
            nml_str = buffer.getvalue().decode("utf-8")
        zipfile.writestr(self.name + ".nml", nml_str)

        for volume_layer in self._volume_layers:
            if volume_layer.zip is None:
                with BytesIO() as buffer:
                    with ZipFile(buffer, mode="a"):
                        pass
                    layer_content = buffer.getvalue()
            else:
                layer_content = volume_layer.zip.read_bytes()
            zipfile.writestr(
                volume_layer._default_zip_name(),
                layer_content,
            )

    def get_remote_annotation_dataset(self) -> Dataset:
        """
        Returns a streamed dataset of the annotation as shown in webknossos,
        incorporating fallback layers and potentially mappings.
        A mapping is currently only incorporated if it is a pinned agglomerate mapping.
        After an agglomerate mapping was activated in WEBKNOSSOS, it is pinned as soon
        as the first volume editing action is done. Note that this behavior might change
        in the future.
        """
        from ..client.context import _get_context

        if self.annotation_id is None:
            raise ValueError(
                "The annotation_id is not set, cannot get the corresponding dataset."
            )

        context = _get_context()
        token: Optional[str]
        if self.organization_id is None:
            token = context.required_token
            organization_id = context.organization_id
        else:
            token = context.token
            organization_id = self.organization_id
            # avoid requiring authentication
            if token is not None:
                if organization_id != context.organization_id:
                    warnings.warn(
                        "[WARNING] The annotation used with `get_remote_annotation_dataset` "
                        + "specifies a different organization id than the current context. "
                        + f"The annotation uses {organization_id}, the context {context.organization_id}.",
                        UserWarning,
                    )

        dataset_info = context.api_client.dataset_info(
            organization_id, self.dataset_name
        )

        datastore_url = dataset_info.data_store.url

        zarr_path = UPath(
            f"{datastore_url}/data/annotations/zarr/{self.annotation_id}/",
            headers={} if token is None else {"X-Auth-Token": token},
        )
        return Dataset.open(zarr_path)

    def get_remote_base_dataset(
        self,
        sharing_token: Optional[str] = None,
        webknossos_url: Optional[str] = None,
    ) -> RemoteDataset:
        return Dataset.open_remote(
            self.dataset_name,
            self.organization_id,
            sharing_token=sharing_token,
            webknossos_url=webknossos_url,
        )

    def get_volume_layer_names(self) -> Iterable[str]:
        return (i.name for i in self._volume_layers)

    def add_volume_layer(
        self,
        name: str,
        fallback_layer: Union[Layer, str, None] = None,
        volume_layer_id: Optional[int] = None,
    ) -> None:
        """
        Adds a volume layer to the annotation, without manual annotations but possibly referring to
        segmentation data using the `fallback_layer`.
        To prepare volume annotations in the code for correction of segmentation data in the WEBKNOSSOS interface,
        please use the `fallback_layer` argument, referencing a segmentation layer that is available on WEBKNOSSOS
        (e.g. using the `Dataset` upload before).
        Correcting segmentations using fallback layers is much more efficient, adding volume annotation data
        programmatically is discouraged therefore."""
        if volume_layer_id is None:
            volume_layer_id = max((i.id for i in self._volume_layers), default=-1) + 1
        else:
            assert (
                volume_layer_id not in [i.id for i in self._volume_layers]
            ), f"volume layer id {volume_layer_id} already exists in annotation {self.name}."
        fallback_layer_name: Optional[str]
        if isinstance(fallback_layer, Layer):
            assert (
                fallback_layer.category == SEGMENTATION_CATEGORY
            ), "The fallback layer must be a segmentation layer."
            fallback_layer_name = fallback_layer.name
        elif fallback_layer is not None:
            fallback_layer_name = str(fallback_layer)
        else:
            fallback_layer_name = None
        self._volume_layers.append(
            _VolumeLayer(
                id=volume_layer_id,
                name=name,
                fallback_layer_name=fallback_layer_name,
                data_format=DataFormat.Zarr3,
                zip=None,
                segments={},
                largest_segment_id=None,
            )
        )

    def _get_volume_layer(
        self,
        volume_layer_name: Optional[str] = None,
        volume_layer_id: Optional[int] = None,
    ) -> _VolumeLayer:
        assert len(self._volume_layers) > 0, "No volume annotations present."

        if len(self._volume_layers) == 1:
            volume_layer = self._volume_layers[0]
            if volume_layer_id is not None and volume_layer_id != volume_layer.id:
                warnings.warn(
                    f"[INFO] Only a single volume annotation is present and its id {volume_layer.id} does not fit the given id {volume_layer_id}."
                )
            if (
                volume_layer_name is not None
                and volume_layer.name is not None
                and volume_layer_name != volume_layer.name
            ):
                warnings.warn(
                    f"[INFO] Only a single volume annotation is present and its name {volume_layer.name} "
                    + f"does not fit the given name {volume_layer_name}."
                )
            return volume_layer

        if volume_layer_id is not None:
            for volume_layer in self._volume_layers:
                if volume_layer_id == volume_layer.id:
                    if (
                        volume_layer_name is not None
                        and volume_layer.name is not None
                        and volume_layer_name != volume_layer.name
                    ):
                        warnings.warn(
                            f"[WARNING] The volume annotation was matched by id {volume_layer_id}, "
                            + f"but its name {volume_layer.name} does not fit the given name {volume_layer_name}."
                        )
                    return volume_layer
            available_ids = [volume_layer.id for volume_layer in self._volume_layers]
            raise ValueError(
                f"Couldn't find a volume annotation with the id {volume_layer_id}, available are {available_ids}."
            )
        elif volume_layer_name is not None:
            fitting_volume_layers = [
                i for i in self._volume_layers if i.name == volume_layer_name
            ]
            assert (
                len(fitting_volume_layers) != 0
            ), f"The specified volume name {volume_layer_name} could not be found in this annotation."
            assert len(fitting_volume_layers) == 1, (
                f"More than one volume annotation has the name {volume_layer_name}. "
                + "Please specify the exact annotation via the volume_layer_id argument. "
                + f"The matching annotations have the ids {[i.id for i in fitting_volume_layers]}"
            )
            return fitting_volume_layers[0]
        else:
            raise ValueError(
                "The annotation contains multiple volume layers. "
                + "Please specify which layer should be used via volume_layer_name or volume_layer_id."
            )

    def delete_volume_layer(
        self,
        volume_layer_name: Optional[str] = None,
        volume_layer_id: Optional[int] = None,
    ) -> None:
        layer_id = self._get_volume_layer(
            volume_layer_name=volume_layer_name,
            volume_layer_id=volume_layer_id,
        ).id
        self._volume_layers = [i for i in self._volume_layers if i.id != layer_id]

    def export_volume_layer_to_dataset(
        self,
        dataset: Dataset,
        layer_name: str = "volume_layer",
        volume_layer_name: Optional[str] = None,
        volume_layer_id: Optional[int] = None,
    ) -> SegmentationLayer:
        """
        Given a dataset, this method will export the specified
        volume annotation of this annotation into that dataset
        by creating a new layer.
        The largest_segment_id is computed automatically, unless provided
        explicitly.

        `volume_layer_name` or `volume_layer_id` has to be provided,
        if the annotation contains multiple volume layers.
        Use `get_volume_layer_names()` to look up available layers.
        """
        volume_layer = self._get_volume_layer(
            volume_layer_name=volume_layer_name,
            volume_layer_id=volume_layer_id,
        )
        volume_zip_path = volume_layer.zip

        largest_segment_id = volume_layer.largest_segment_id

        assert (
            volume_zip_path is not None
        ), "The selected volume layer data is not available and cannot be exported."

        with volume_zip_path.open(mode="rb") as f:
            data_zip = ZipFile(f)
            if volume_layer.data_format == DataFormat.WKW:
                wrong_files = [
                    i.filename
                    for i in data_zip.filelist
                    if ANNOTATION_WKW_PATH_RE.search(i.filename) is None
                ]
                assert (
                    len(wrong_files) == 0
                ), f"The annotation contains unexpected files: {wrong_files}"
                data_zip.extractall(dataset.path / layer_name)
                layer = cast(
                    SegmentationLayer,
                    dataset.add_layer_for_existing_files(
                        layer_name,
                        category=SEGMENTATION_CATEGORY,
                        largest_segment_id=largest_segment_id,
                    ),
                )
            elif volume_layer.data_format == DataFormat.Zarr3:
                datasource_properties = dataset_converter.structure(
                    json.loads(data_zip.read(PROPERTIES_FILE_NAME)), DatasetProperties
                )
                assert (
                    len(datasource_properties.data_layers) == 1
                ), f"Volume data zip must contain exactly one layer, got {len(datasource_properties.data_layers)}"
                layer_properties = datasource_properties.data_layers[0]
                internal_layer_name = layer_properties.name
                layer_properties.name = layer_name

                _extract_zip_folder(
                    data_zip, dataset.path / layer_name, f"{internal_layer_name}/"
                )

                layer = cast(
                    SegmentationLayer,
                    dataset._add_existing_layer(layer_properties),
                )

        best_mag_view = layer.get_finest_mag()

        if largest_segment_id is None:
            max_value = max(
                (
                    view.read().max()
                    for view in best_mag_view.get_views_on_disk(read_only=True)
                ),
                default=0,
            )
            layer.largest_segment_id = int(max_value)
        else:
            layer.largest_segment_id = largest_segment_id
        return layer

    @contextmanager
    def temporary_volume_layer_copy(
        self,
        volume_layer_name: Optional[str] = None,
        volume_layer_id: Optional[int] = None,
        read_only: bool = True,
    ) -> Iterator[SegmentationLayer]:
        """
        Given a volume annotation path, create a temporary dataset which
        contains the volume annotation. Returns the corresponding `Layer`.

        `volume_layer_name` or `volume_layer_id` has to be provided,
        if the annotation contains multiple volume layers.
        """

        with TemporaryDirectory() as tmp_annotation_dir:
            input_annotation_dataset = Dataset(
                tmp_annotation_dir,
                name="tmp_annotation_dataset",
                voxel_size=self.voxel_size,
                exist_ok=True,
            )

            input_annotation_layer = self.export_volume_layer_to_dataset(
                input_annotation_dataset,
                "volume_layer",
                volume_layer_name=volume_layer_name,
                volume_layer_id=volume_layer_id,
            )

            input_annotation_dataset._read_only = read_only

            yield input_annotation_layer

    def get_volume_layer_segments(
        self,
        volume_layer_name: Optional[str] = None,
        volume_layer_id: Optional[int] = None,
    ) -> Dict[int, SegmentInformation]:
        """Returns a dict mapping from segment ids to `SegmentInformation`.
        The dict is mutable, changes to the returned instance are saved in the local annotation.
        Changes in a downloaded annotation that are done online in webknossos are not
        reflected automatically, the annotation needs to be re-downloaded."""
        layer = self._get_volume_layer(
            volume_layer_name=volume_layer_name,
            volume_layer_id=volume_layer_id,
        )
        return layer.segments


Annotation._set_init_docstring()


@unique
class AnnotationType(Enum):
    """Annotations can be of different types which has to be specified when using `Annotation.download()`
    with an annotation id."""

    EXPLORATIONAL = "Explorational"
    """**Explorational** annotations are all annotations created without the task system, e.g.
    by uploading an annotation or using the "Create Annotation" Button in the dataset view in webknossos."""
    TASK = "Task"
    """The **Task** type is automatically assigned to all annotations that are instances of a task.
    See also `Task`."""
    _COMPOUND_TASK = "CompoundTask"
    """A collection of annotations of all instances of a task. Currently not supported,
    please download all annotations individually instead, see `Task`."""
    _COMPOUND_PROJECT = "CompoundProject"
    """A collection of annotations of all instances of all tasks in a project. Currently not supported,
    please download all annotations individually instead, see `Project`."""
    _COMPOUND_TASK_TYPE = "CompoundTaskType"
    """A collection of annotations of all instances of all tasks in a project. Currently not supported,
    please download all annotations individually instead, see `Project`."""


@unique
class AnnotationState(Enum):
    """This Enum contains the state of annotations belonging to tasks.
    Can be retrieved via `Task` instances, getting `AnnotationInfo` from `task.get_annotation_infos()`.
    """

    FINISHED = "Finished"
    ACTIVE = "Active"
    CANCELLED = "Cancelled"
    INITIALIZING = "Initializing"


_COMPOUND_ANNOTATION_TYPES = [
    AnnotationType._COMPOUND_PROJECT,
    AnnotationType._COMPOUND_TASK,
    AnnotationType._COMPOUND_TASK_TYPE,
]

_ANNOTATION_URL_REGEX = re.compile(
    r"^(?P<webknossos_url>https?://.*)/annotations/"
    + rf"((?P<annotation_type>{'|'.join(i.value for i in AnnotationType.__members__.values())})/)?"
    + r"(?P<annotation_id>[0-9A-Fa-f]*)"
)


def open_annotation(annotation_path: Union[str, PathLike]) -> "Annotation":
    """Deprecated."""
    if Path(annotation_path).exists():
        warn_deprecated("open_annotation", "Annotation.load")
        return Annotation.load(annotation_path)
    else:
        assert isinstance(
            annotation_path, str
        ), f"Called open_annotation with a path-like, but {annotation_path} does not exist."
        warn_deprecated("open_annotation", "Annotation.download")
        return Annotation.download(annotation_path)


def _parse_filename_from_header(value: str) -> str:
    # Adapted from https://peps.python.org/pep-0594/#cgi
    from email.message import Message

    m = Message()
    m["content-type"] = value
    return dict(m.get_params() or []).get("filename", "")



import logging
import re
from urllib.parse import urlparse

from .context import _get_api_client, webknossos_context

logger = logging.getLogger(__name__)


_SHORT_LINK_REGEX = re.compile(r"^/links/(?P<short_link_key>.+)$")


def resolve_short_link(url: str) -> str:
    try:
        parts = urlparse(url)
        if parts.scheme not in ["http", "https"]:
            return url
        if parts.netloc == "":
            return url

        match = re.match(_SHORT_LINK_REGEX, parts.path)
        if match is None:
            return url

        webknossos_url = f"{parts.scheme}://{parts.netloc}"
        short_link_key = match.group("short_link_key")
        with webknossos_context(url=webknossos_url):
            client = _get_api_client()
            return client.short_link_by_key(key=short_link_key).long_link
    except Exception as e:
        logger.warning(f"Got an error during short link resolution of link {url}: {e}")
        return url



from contextlib import contextmanager
from os import PathLike
from pathlib import Path
from typing import Dict, Iterator, List, Set

import numpy as np

try:
    from pims import FramesSequenceND
except ImportError as e:
    raise RuntimeError(
        "Cannot import pims, please install it e.g. using 'webknossos[all]'"
    ) from e

try:
    from pylibCZIrw import czi as pyczi
except ImportError as e:
    raise RuntimeError(
        "Cannot import pylibCZIrw, please install it e.g. using 'webknossos[czi]'"
    ) from e

PIXEL_TYPE_TO_DTYPE = {
    "Gray8": "<u1",
    "Gray16": "<u2",
    "Gray32Float": "<f4",
    "Bgr24": "<u1",
    "Bgr48": "<u2",
    "Bgr96Float": "<f4",
    "Bgra32": "<4u1",
    "Gray64ComplexFloat": "<F8",
    "Bgr192ComplexFloat": "<F8",
    "Gray32": "<i4",
    "Gray64": "<i8",
}


class PimsCziReader(FramesSequenceND):
    @classmethod
    def class_exts(cls) -> Set[str]:
        return {"czi"}

    # class_priority is used in pims to pick the reader with the highest priority.
    # Default is 10, and bioformats priority (which is the only other reader supporting czi) is 2.
    # See http://soft-matter.github.io/pims/v0.6.1/custom_readers.html#plugging-into-pims-s-open-function
    class_priority = 20

    def __init__(self, path: PathLike, czi_channel: int = 0) -> None:
        self.path = Path(path)
        self.czi_channel = czi_channel
        self.axis_offsets: Dict[str, int] = {}
        super().__init__()
        with self.czi_file() as czi_file:
            for axis, (
                start,
                end,
            ) in czi_file.total_bounding_box.items():
                axis = axis.lower()
                if axis == "c":
                    continue
                length = end - start
                if axis not in "xy" and length == 1:
                    # not propagating axes of length one
                    continue
                assert length >= 0, f"axis length must be >= 0, got {length}"
                self._init_axis(axis, length)
                self.axis_offsets[axis] = start
            self._czi_pixel_type = czi_file.get_channel_pixel_type(self.czi_channel)
            if self._czi_pixel_type.startswith("Bgra"):
                self._init_axis("c", 4)
            elif self._czi_pixel_type.startswith("Bgr"):
                self._init_axis("c", 3)
            elif self._czi_pixel_type.startswith("Gray"):
                self._init_axis("c", 1)
            elif self._czi_pixel_type == "Invalid":
                raise ValueError(
                    f"czi_channel {self.czi_channel} does not exist in {self.path}"
                )
            else:
                raise ValueError(
                    f"Got unsupported czi pixel-type {self._czi_pixel_type} in {self.path}"
                )

        self._register_get_frame(self.get_frame_2D, "yxc")

    @contextmanager
    def czi_file(self) -> Iterator[pyczi.CziReader]:
        with pyczi.open_czi(str(self.path)) as czi_file:
            yield czi_file

    def available_czi_channels(self) -> List[int]:
        with self.czi_file() as czi_file:
            return sorted(czi_file.pixel_types.keys())

    @property  # potential @cached_property for py3.8+
    def pixel_type(self) -> np.dtype:
        return np.dtype(PIXEL_TYPE_TO_DTYPE[self._czi_pixel_type])

    def get_frame_2D(self, **ind: int) -> np.ndarray:
        plane = {k.upper(): v for k, v in ind.items()}
        for axis in plane.keys():
            if axis in self.axis_offsets:
                plane[axis] += self.axis_offsets[axis]

        # safe-guard against x/y in ind argument,
        # we always read the whole slice here:
        plane.pop("X", None)
        plane.pop("Y", None)

        plane["C"] = self.czi_channel
        with self.czi_file() as czi_file:
            a = czi_file.read(plane=plane)
            num_channels = a.shape[-1]
            if num_channels == 3:
                # convert from bgr to rgb
                a = np.flip(a, axis=-1)
            elif num_channels == 4:
                # convert from bgra to rgba
                a_red = a[:, :, 2].copy(order="K")
                a[:, :, 2] = a[:, :, 0]
                a[:, :, 0] = a_red
            return a



import warnings
from argparse import Namespace
from pathlib import Path
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
)

import numpy as np
import wkw
from cluster_tools import Executor

from webknossos.geometry.vec_int import VecInt

from ..geometry import BoundingBox, Mag, NDBoundingBox, Vec3Int, Vec3IntLike
from ..utils import (
    get_executor_for_args,
    get_rich_progress,
    wait_and_ensure_success,
    warn_deprecated,
)
from ._array import ArrayInfo, BaseArray, WKWArray

if TYPE_CHECKING:
    from ._utils.buffered_slice_reader import BufferedSliceReader
    from ._utils.buffered_slice_writer import BufferedSliceWriter


def _assert_check_equality(args: Tuple["View", "View", int]) -> None:
    view_a, view_b, _ = args
    assert np.all(view_a.read() == view_b.read())


_BLOCK_ALIGNMENT_WARNING = (
    "[WARNING] write() was called on a compressed mag without block alignment. "
    + "Performance will be degraded as the data has to be padded first."
)


class View:
    """
    A `View` is essentially a bounding box to a region of a specific `StorageBackend` that also provides functionality.
    Write-operations are restricted to the bounding box.
    `View`s are designed to be easily passed around as parameters.
    A `View`, in its most basic form, does not have a reference to its `StorageBackend`.
    """

    _path: Path
    _array_info: ArrayInfo
    _bounding_box: Optional[NDBoundingBox]
    _read_only: bool
    _cached_array: Optional[BaseArray]
    _mag: Mag

    def __init__(
        self,
        path_to_mag_view: Path,
        array_info: ArrayInfo,
        bounding_box: Optional[
            NDBoundingBox
        ],  # in mag 1, absolute coordinates, optional only for mag_view since it overwrites the bounding_box property
        mag: Mag,
        read_only: bool = False,
    ):
        """
        Do not use this constructor manually. Instead use `View.get_view()` (also available on a `MagView`) to get a `View`.
        """
        self._path = path_to_mag_view
        self._array_info = array_info
        self._bounding_box = bounding_box
        self._read_only = read_only
        self._cached_array = None
        self._mag = mag

    @property
    def info(self) -> ArrayInfo:
        return self._array_info

    @property
    def header(self) -> wkw.Header:
        """⚠️ Deprecated, use `info` instead."""
        warn_deprecated("header", "info")
        assert isinstance(
            self._array, WKWArray
        ), "`header` only works with WKW datasets."
        return self._array._wkw_dataset.header

    @property
    def bounding_box(self) -> NDBoundingBox:
        assert self._bounding_box is not None
        return self._bounding_box

    @property
    def mag(self) -> Mag:
        return self._mag

    @property
    def read_only(self) -> bool:
        return self._read_only

    @property
    def global_offset(self) -> VecInt:
        """⚠️ Deprecated, use `view.bounding_box.in_mag(view.mag).topleft` instead."""
        warnings.warn(
            "[DEPRECATION] view.global_offset is deprecated. "
            + "Since this is a View, please use "
            + "view.bounding_box.in_mag(view.mag).topleft instead.",
            DeprecationWarning,
        )
        return self.bounding_box.in_mag(self._mag).topleft

    @property
    def size(self) -> VecInt:
        """⚠️ Deprecated, use `view.bounding_box.in_mag(view.mag).size` instead."""
        warnings.warn(
            "[DEPRECATION] view.size is deprecated. "
            + "Since this is a View, please use "
            + "view.bounding_box.in_mag(view.mag).size instead.",
            DeprecationWarning,
        )
        return self.bounding_box.in_mag(self._mag).size

    def _get_mag1_bbox(
        self,
        abs_mag1_bbox: Optional[NDBoundingBox] = None,
        rel_mag1_bbox: Optional[NDBoundingBox] = None,
        abs_mag1_offset: Optional[Vec3IntLike] = None,
        rel_mag1_offset: Optional[Vec3IntLike] = None,
        mag1_size: Optional[Vec3IntLike] = None,
        abs_current_mag_offset: Optional[Vec3IntLike] = None,
        rel_current_mag_offset: Optional[Vec3IntLike] = None,
        current_mag_size: Optional[Vec3IntLike] = None,
    ) -> NDBoundingBox:
        num_bboxes = _count_defined_values([abs_mag1_bbox, rel_mag1_bbox])
        num_offsets = _count_defined_values(
            [
                abs_mag1_offset,
                rel_mag1_offset,
                abs_current_mag_offset,
                rel_current_mag_offset,
            ]
        )
        num_sizes = _count_defined_values([mag1_size, current_mag_size])
        if num_bboxes == 0:
            assert num_offsets != 0, "You must supply an offset or a bounding box."
            assert (
                num_sizes != 0
            ), "When supplying an offset, you must also supply a size. Alternatively, supply a bounding box."
            assert num_offsets == 1, "Only one offset can be supplied."
            assert num_sizes == 1, "Only one size can be supplied."
        else:
            assert num_bboxes == 1, "Only one bounding-box can be supplied."
            assert (
                num_offsets == 0
            ), "A bounding-box was supplied, you cannot also supply an offset."
            assert (
                num_sizes == 0
            ), "A bounding-box was supplied, you cannot also supply a size."

        if abs_mag1_bbox is not None:
            return abs_mag1_bbox

        if rel_mag1_bbox is not None:
            return rel_mag1_bbox.offset(self.bounding_box.topleft)

        else:
            mag_vec = self._mag.to_vec3_int()
            if rel_current_mag_offset is not None:
                abs_mag1_offset = (
                    self.bounding_box.topleft
                    + Vec3Int(rel_current_mag_offset) * mag_vec
                )
            if abs_current_mag_offset is not None:
                abs_mag1_offset = Vec3Int(abs_current_mag_offset) * mag_vec
            if rel_mag1_offset is not None:
                abs_mag1_offset = self.bounding_box.topleft + rel_mag1_offset

            if current_mag_size is not None:
                mag1_size = Vec3Int(current_mag_size) * mag_vec

            assert abs_mag1_offset is not None, "No offset was supplied."
            assert mag1_size is not None, "No size was supplied."

            assert (
                len(self.bounding_box) == 3
            ), "The delivered offset and size are only usable for 3D views."

            return self.bounding_box.with_topleft(abs_mag1_offset).with_size(mag1_size)

    def write(
        self,
        data: np.ndarray,
        offset: Optional[Vec3IntLike] = None,  # deprecated, relative, in current mag
        json_update_allowed: bool = True,
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
    ) -> None:
        """
        The user can specify where the data should be written.
        The default is to write the data to the view's bounding box.
        Alternatively, one can supply one of the following keywords:
        * `relative_offset` in Mag(1) -> only usable for 3D datasets
        * `absolute_offset` in Mag(1) -> only usable for 3D datasets
        * `relative_bounding_box` in Mag(1)
        * `absolute_bounding_box` in Mag(1)

        ⚠️ The `offset` parameter is deprecated.
        This parameter used to be relative for `View` and absolute for `MagView`,
        and specified in the mag of the respective view.

        Writing data to a segmentation layer manually does not automatically update the largest_segment_id. To set
        the largest segment id properly run the `refresh_largest_segment_id` method on your layer or set the
        `largest_segment_id` property manually..

        Example:

        ```python
        ds = Dataset(DS_PATH, voxel_size=(1, 1, 1))

        segmentation_layer = cast(
            SegmentationLayer,
            ds.add_layer("segmentation", SEGMENTATION_CATEGORY),
        )
        mag = segmentation_layer.add_mag(Mag(1))

        mag.write(data=MY_NP_ARRAY)

        segmentation_layer.refresh_largest_segment_id()
        ```

        Note that writing compressed data which is not aligned with the blocks on disk may result in
        diminished performance, as full blocks will automatically be read to pad the write actions.
        """
        assert not self.read_only, "Cannot write data to an read_only View"

        if all(
            i is None
            for i in [
                offset,
                absolute_offset,
                relative_offset,
                absolute_bounding_box,
                relative_bounding_box,
            ]
        ):
            relative_offset = Vec3Int.zeros()

        if (absolute_bounding_box or relative_bounding_box) is not None:
            data_shape = None
        else:
            data_shape = Vec3Int(data.shape[-3:])

        if offset is not None:
            if self._mag == Mag(1):
                alternative = "Since this is a View in Mag(1), please use view.write(relative_offset=my_vec)"
            else:
                alternative = (
                    "Since this is a View, please use the coordinates in Mag(1) instead, e.g. "
                    + "view.write(relative_offset=my_vec * view.mag.to_vec3_int())"
                )

            warnings.warn(
                "[DEPRECATION] Using view.write(offset=my_vec) is deprecated. "
                + "Please use relative_offset or absolute_offset instead. "
                + alternative,
                DeprecationWarning,
            )

        num_channels = self._array_info.num_channels
        if len(data.shape) == 3:
            assert (
                num_channels == 1
            ), f"The number of channels of the dataset ({num_channels}) does not match the number of channels of the passed data (1)"
        else:
            assert (
                num_channels == data.shape[0]
            ), f"The number of channels of the dataset ({num_channels}) does not match the number of channels of the passed data ({data.shape[0]})"

        mag1_bbox = self._get_mag1_bbox(
            rel_current_mag_offset=offset,
            rel_mag1_offset=relative_offset,
            abs_mag1_offset=absolute_offset,
            rel_mag1_bbox=relative_bounding_box,
            abs_mag1_bbox=absolute_bounding_box,
            current_mag_size=data_shape,
        )
        if json_update_allowed:
            assert self.bounding_box.contains_bbox(
                mag1_bbox
            ), f"The bounding box to write {mag1_bbox} is larger than the view's bounding box {self.bounding_box}"

        current_mag_bbox = mag1_bbox.in_mag(self._mag)

        if self._is_compressed():
            for current_mag_bbox, chunked_data in self._prepare_compressed_write(
                current_mag_bbox, data, json_update_allowed
            ):
                self._array.write(current_mag_bbox, chunked_data)
        else:
            self._array.write(current_mag_bbox, data)

    def _prepare_compressed_write(
        self,
        current_mag_bbox: NDBoundingBox,
        data: np.ndarray,
        json_update_allowed: bool = True,
    ) -> Iterator[Tuple[NDBoundingBox, np.ndarray]]:
        """This method takes an arbitrary sized chunk of data with an accompanying bbox,
        divides these into chunks of shard_shape size and delegates
        the preparation to _prepare_compressed_write_chunk."""

        chunked_bboxes = current_mag_bbox.chunk(
            self.info.shard_shape,
            chunk_border_alignments=self.info.shard_shape,
        )
        for chunked_bbox in chunked_bboxes:
            source_slice: Any
            if len(data.shape) == 3:
                source_slice = chunked_bbox.offset(
                    -current_mag_bbox.topleft
                ).to_slices()
            else:
                source_slice = (slice(None, None),) + chunked_bbox.offset(
                    -current_mag_bbox.topleft
                ).to_slices()

            yield self._prepare_compressed_write_chunk(
                chunked_bbox, data[source_slice], json_update_allowed
            )

    def _prepare_compressed_write_chunk(
        self,
        current_mag_bbox: NDBoundingBox,
        data: np.ndarray,
        json_update_allowed: bool = True,
    ) -> Tuple[NDBoundingBox, np.ndarray]:
        """This method takes an arbitrary sized chunk of data with an accompanying bbox
        (ideally not larger than a shard) and enlarges that chunk to fit the shard it
        resides in (by reading the entire shard data and writing the passed data ndarray
        into the specified volume). That way, the returned data can be written as a whole
        shard which is a requirement for compressed writes."""

        aligned_bbox = current_mag_bbox.align_with_mag(self.info.shard_shape, ceil=True)

        if current_mag_bbox != aligned_bbox:
            # The data bbox should either be aligned or match the dataset's bounding box:
            current_mag_view_bbox = self.bounding_box.in_mag(self._mag)
            if (
                json_update_allowed
                and current_mag_bbox
                != current_mag_view_bbox.intersected_with(aligned_bbox)
            ):
                warnings.warn(
                    _BLOCK_ALIGNMENT_WARNING,
                )

            aligned_data = self._read_without_checks(aligned_bbox)

            index_slice = (slice(None, None),) + current_mag_bbox.offset(
                -aligned_bbox.topleft
            ).to_slices()
            # overwrite the specified data
            aligned_data[index_slice] = data
            return aligned_bbox, aligned_data

        return current_mag_bbox, data

    def read(
        self,
        offset: Optional[Vec3IntLike] = None,  # deprecated, relative, in current mag
        size: Optional[
            Vec3IntLike
        ] = None,  # usually in mag1, in current mag if offset is given
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
    ) -> np.ndarray:
        """
        The user can specify which data should be read.
        The default is to read all data of the view's bounding box.
        Alternatively, one can supply one of the following keyword argument combinations:
        * `relative_offset` and `size`, both in Mag(1)
        * `absolute_offset` and `size`, both in Mag(1)
        * `relative_bounding_box` in Mag(1)
        * `absolute_bounding_box` in Mag(1)
        * ⚠️ deprecated: `offset` and `size`, both in the current Mag.
          `offset` used to be relative for `View` and absolute for `MagView`

        If the specified bounding box exceeds the data on disk, the rest is padded with `0`.

        Returns the specified data as a `np.array`.


        Example:
        ```python
        import numpy as np

        # ...
        # let mag1 be a MagView
        view = mag1.get_view(absolute_offset=(10, 20, 30), size=(100, 200, 300))

        assert np.array_equal(
            view.read(absolute_offset=(0, 0, 0), size=(100, 200, 300)),
            view.read(),
        )
        ```
        """

        current_mag_size: Optional[Vec3IntLike]
        mag1_size: Optional[Vec3IntLike]
        if absolute_bounding_box is None and relative_bounding_box is None:
            if offset is None:
                if size is None:
                    assert (
                        relative_offset is None and absolute_offset is None
                    ), "You must supply size, when reading with an offset."
                    absolute_bounding_box = self.bounding_box
                    current_mag_size = None
                    mag1_size = None
                else:
                    if relative_offset is None and absolute_offset is None:
                        if type(self) == View:
                            offset_param = "relative_offset"
                        else:
                            offset_param = "absolute_offset"
                        warnings.warn(
                            "[DEPRECATION] Using view.read(size=my_vec) only with a size is deprecated. "
                            + f"Please use view.read({offset_param}=(0, 0, 0), size=size_vec * view.mag.to_vec3_int()) instead.",
                            DeprecationWarning,
                        )
                        current_mag_size = size
                        mag1_size = None
                    else:
                        current_mag_size = None
                        mag1_size = size
            else:
                view_class = type(self).__name__
                if type(self) == View:
                    offset_param = "relative_offset"
                else:
                    offset_param = "absolute_offset"
                if self._mag == Mag(1):
                    alternative = f"Since this is a {view_class} in Mag(1), please use view.read({offset_param}=my_vec, size=size_vec)"
                else:
                    alternative = (
                        f"Since this is a {view_class}, please use the coordinates in Mag(1) instead, e.g. "
                        + f"view.read({offset_param}=my_vec * view.mag.to_vec3_int(),  size=size_vec * view.mag.to_vec3_int())"
                    )

                warnings.warn(
                    "[DEPRECATION] Using view.read(offset=my_vec) is deprecated. "
                    + "Please use relative_offset or absolute_offset instead. "
                    + alternative,
                    DeprecationWarning,
                )

                if size is None:
                    absolute_bounding_box = self.bounding_box.offset(
                        self._mag.to_vec3_int() * offset
                    )
                    offset = None
                    current_mag_size = None
                    mag1_size = None
                else:
                    # (deprecated) offset and size are given
                    current_mag_size = size
                    mag1_size = None

            if all(
                i is None
                for i in [
                    offset,
                    absolute_offset,
                    relative_offset,
                    absolute_bounding_box,
                ]
            ):
                relative_offset = Vec3Int.zeros()
        else:
            assert (
                size is None
            ), "Cannot supply a size when using bounding_box in view.read()"
            # offset is asserted anyways in _get_mag1_bbox
            current_mag_size = None
            mag1_size = None

        mag1_bbox = self._get_mag1_bbox(
            rel_current_mag_offset=offset,
            rel_mag1_offset=relative_offset,
            abs_mag1_offset=absolute_offset,
            current_mag_size=current_mag_size,
            mag1_size=mag1_size,
            abs_mag1_bbox=absolute_bounding_box,
            rel_mag1_bbox=relative_bounding_box,
        )
        assert not mag1_bbox.is_empty(), (
            f"The size ({mag1_bbox.size} in mag1) contains a zero. "
            + "All dimensions must be strictly larger than '0'."
        )
        assert mag1_bbox.topleft.is_positive(), f"The offset ({mag1_bbox.topleft} in mag1) must not contain negative dimensions."

        return self._read_without_checks(mag1_bbox.in_mag(self._mag))

    def read_bbox(self, bounding_box: Optional[BoundingBox] = None) -> np.ndarray:
        """
        ⚠️ Deprecated. Please use `read()` with `relative_bounding_box` or `absolute_bounding_box` in Mag(1) instead.
        The user can specify the `bounding_box` in the current mag of the requested data.
        See `read()` for more details.
        """

        view_class = type(self).__name__
        if type(self) == View:
            offset_param = "relative_bounding_box"
        else:
            offset_param = "absolute_bounding_box"
        if self._mag == Mag(1):
            alternative = f"Since this is a {view_class} in Mag(1), please use view.read({offset_param}=bbox)"
        else:
            alternative = (
                f"Since this is a {view_class}, please use the bbox in Mag(1) instead, e.g. "
                + f"view.read({offset_param}=bbox.from_mag_to_mag1(view.mag))"
            )

        warnings.warn(
            "[DEPRECATION] read_bbox() (with a bbox in the current mag) is deprecated. "
            + "Please use read() with relative_bounding_box or absolute_bounding_box in Mag(1) instead. "
            + alternative,
            DeprecationWarning,
        )
        if bounding_box is None:
            return self.read()
        else:
            return self.read(bounding_box.topleft, bounding_box.size)

    def _read_without_checks(
        self,
        current_mag_bbox: NDBoundingBox,
    ) -> np.ndarray:
        data = self._array.read(current_mag_bbox)
        return data

    def get_view(
        self,
        offset: Optional[Vec3IntLike] = None,
        size: Optional[Vec3IntLike] = None,
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bbox: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bbox: Optional[NDBoundingBox] = None,  # in mag1
        read_only: Optional[bool] = None,
    ) -> "View":
        """
        Returns a view that is limited to the specified bounding box.
        The new view may exceed the bounding box of the current view only if `read_only` is set to `True`.

        The default is to return the same view as the current bounding box,
        in case of a `MagView` that's the layer bounding box.
        One can supply one of the following keyword argument combinations:
        * `relative_offset` and `size`, both in Mag(1)
        * `absolute_offset` and `size`, both in Mag(1)
        * ⚠️ deprecated: `offset` and `size`, both in the current Mag.
          `offset` used to be relative for `View` and absolute for `MagView`

        Example:
        ```python
        # ...
        # let mag1 be a MagView
        view = mag1.get_view(absolute_offset=(10, 20, 30), size=(100, 200, 300))

        # works because the specified sub-view is completely in the bounding box of the view
        sub_view = view.get_view(relative_offset=(50, 60, 70), size=(10, 120, 230))

        # fails because the specified sub-view is not completely in the bounding box of the view
        invalid_sub_view = view.get_view(relative_offset=(50, 60, 70), size=(999, 120, 230))

        # works because `read_only=True`
        valid_sub_view = view.get_view(relative_offset=(50, 60, 70), size=(999, 120, 230), read_only=True)
        ```
        """
        if read_only is None:
            read_only = self.read_only
        else:
            assert (
                read_only or not self.read_only
            ), "Failed to get subview. The calling view is read_only. Therefore, the subview also has to be read_only."

        current_mag_size: Optional[Vec3IntLike]
        mag1_size: Optional[Vec3IntLike]

        if offset is None:
            if size is None:
                assert (
                    relative_offset is None and absolute_offset is None
                ), "You must supply a size, when using get_view with an offset."
                current_mag_size = None
                mag1_size = self.bounding_box.size
            else:
                if relative_offset is None and absolute_offset is None:
                    if type(self) == View:
                        offset_param = "relative_offset"
                    else:
                        offset_param = "absolute_offset"
                    warnings.warn(
                        "[DEPRECATION] Using view.get_view(size=my_vec) only with a size is deprecated. "
                        + f"Please use view.get_view({offset_param}=(0, 0, 0), size=size_vec * view.mag.to_vec3_int()) instead.",
                        DeprecationWarning,
                    )
                    current_mag_size = size
                    mag1_size = None
                else:
                    current_mag_size = None
                    mag1_size = size
        else:
            view_class = type(self).__name__
            if type(self) == View:
                offset_param = "relative_offset"
            else:
                offset_param = "absolute_offset"
            if self._mag == Mag(1):
                alternative = f"Since this is a {view_class} in Mag(1), please use view.get_view({offset_param}=my_vec, size=size_vec)"
            else:
                alternative = (
                    f"Since this is a {view_class}, please use the coordinates in Mag(1) instead, e.g. "
                    + f"view.get_view({offset_param}=my_vec * view.mag.to_vec3_int(),  size=size_vec * view.mag.to_vec3_int())"
                )

            warnings.warn(
                "[DEPRECATION] Using view.get_view(offset=my_vec) is deprecated. "
                + "Please use relative_offset or absolute_offset instead. "
                + alternative,
                DeprecationWarning,
            )

            if size is None:
                current_mag_size = None
                mag1_size = self.bounding_box.size
            else:
                # (deprecated) offset and size are given
                current_mag_size = size
                mag1_size = None

        if offset is None and relative_offset is None and absolute_offset is None:
            relative_offset = Vec3Int.zeros()

        mag1_bbox = self._get_mag1_bbox(
            abs_mag1_bbox=absolute_bbox,
            rel_mag1_bbox=relative_bbox,
            rel_current_mag_offset=offset,
            rel_mag1_offset=relative_offset,
            abs_mag1_offset=absolute_offset,
            current_mag_size=current_mag_size,
            mag1_size=mag1_size,
        )
        if not self.bounding_box.is_empty():
            assert not mag1_bbox.is_empty(), (
                f"The size ({mag1_bbox.size} in mag1) contains a zero. "
                + "All dimensions must be strictly larger than '0'."
            )
        assert mag1_bbox.topleft.is_positive(), f"The offset ({mag1_bbox.topleft} in mag1) must not contain negative dimensions."

        if not read_only:
            assert self.bounding_box.contains_bbox(mag1_bbox), (
                f"The bounding box of the new subview {mag1_bbox} is larger than the view's bounding box {self.bounding_box}. "
                + "This is only allowed for read-only views."
            )

            current_mag_bbox = mag1_bbox.in_mag(self._mag)
            current_mag_aligned_bbox = current_mag_bbox.align_with_mag(
                self.info.shard_shape, ceil=True
            )
            # The data bbox should either be aligned or match the dataset's bounding box:
            current_mag_view_bbox = self.bounding_box.in_mag(self._mag)
            if current_mag_bbox != current_mag_view_bbox.intersected_with(
                current_mag_aligned_bbox, dont_assert=True
            ):
                warnings.warn(
                    "[WARNING] get_view() was called without block alignment. "
                    + "Please only use sequentially, parallel access across such views is error-prone.",
                    UserWarning,
                )

        return View(
            self._path,
            self.info,
            bounding_box=mag1_bbox,
            mag=self._mag,
            read_only=read_only,
        )

    def get_buffered_slice_writer(
        self,
        offset: Optional[Vec3IntLike] = None,
        buffer_size: int = 32,
        dimension: int = 2,  # z
        # json_update_allowed enables the update of the bounding box and rewriting of the properties json.
        # It should be False when parallel access is intended.
        json_update_allowed: bool = True,
        *,
        relative_offset: Optional[Vec3IntLike] = None,  # in mag1
        absolute_offset: Optional[Vec3IntLike] = None,  # in mag1
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        use_logging: bool = False,
    ) -> "BufferedSliceWriter":
        """
        The returned writer buffers multiple slices before they are written to disk.
        As soon as the buffer is full, the data gets written to disk.

        Arguments:
        * The user can specify where the writer should start:
            * `relative_offset` in Mag(1)
            * `absolute_offset` in Mag(1)
            * `relative_bounding_box` in Mag(1)
            * `absolute_bounding_box` in Mag(1)
            * ⚠️ deprecated: `offset` in the current Mag,
              used to be relative for `View` and absolute for `MagView`
        * `buffer_size`: amount of slices that get buffered
        * `dimension`: dimension along which the data is sliced
          (x: `0`, y: `1`, z: `2`; default is `2`)).

        The writer must be used as context manager using the `with` syntax (see example below),
        which results in a generator consuming np.ndarray-slices via `writer.send(slice)`.
        Exiting the context will automatically flush any remaining buffered data to disk.

        Usage:
        ```python
        data_cube = ...
        view = ...
        with view.get_buffered_slice_writer() as writer:
            for data_slice in data_cube:
                writer.send(data_slice)
        ```
        """
        from ._utils.buffered_slice_writer import BufferedSliceWriter

        assert (
            not self._read_only
        ), "Cannot get a buffered slice writer on a read-only view."

        return BufferedSliceWriter(
            view=self,
            offset=offset,
            json_update_allowed=json_update_allowed,
            buffer_size=buffer_size,
            dimension=dimension,
            relative_offset=relative_offset,
            absolute_offset=absolute_offset,
            relative_bounding_box=relative_bounding_box,
            absolute_bounding_box=absolute_bounding_box,
            use_logging=use_logging,
        )

    def get_buffered_slice_reader(
        self,
        offset: Optional[Vec3IntLike] = None,
        size: Optional[Vec3IntLike] = None,
        buffer_size: int = 32,
        dimension: int = 2,  # z
        *,
        relative_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        absolute_bounding_box: Optional[NDBoundingBox] = None,  # in mag1
        use_logging: bool = False,
    ) -> "BufferedSliceReader":
        """
        The returned reader yields slices of data along a specified axis.
        Internally, it reads multiple slices from disk at once and buffers the data.

        Arguments:
        * The user can specify where the writer should start:
            * `relative_bounding_box` in Mag(1)
            * `absolute_bounding_box` in Mag(1)
            * ⚠️ deprecated: `offset` and `size` in the current Mag,
              `offset` used to be relative for `View` and absolute for `MagView`
        * `buffer_size`: amount of slices that get buffered
        * `dimension`: dimension along which the data is sliced
          (x: `0`, y: `1`, z: `2`; default is `2`)).

        The reader must be used as a context manager using the `with` syntax (see example below).
        Entering the context returns an iterator yielding slices (np.ndarray).

        Usage:
        ```python
        view = ...
        with view.get_buffered_slice_reader() as reader:
            for slice_data in reader:
                ...
        ```
        """
        from ._utils.buffered_slice_reader import BufferedSliceReader

        return BufferedSliceReader(
            view=self,
            offset=offset,
            size=size,
            buffer_size=buffer_size,
            dimension=dimension,
            relative_bounding_box=relative_bounding_box,
            absolute_bounding_box=absolute_bounding_box,
            use_logging=use_logging,
        )

    def for_each_chunk(
        self,
        func_per_chunk: Callable[[Tuple["View", int]], None],
        chunk_shape: Optional[Vec3IntLike] = None,  # in Mag(1)
        executor: Optional[Executor] = None,
        progress_desc: Optional[str] = None,
        *,
        chunk_size: Optional[Vec3IntLike] = None,  # deprecated
    ) -> None:
        """
        The view is chunked into multiple sub-views of size `chunk_shape` (in Mag(1)),
        by default one chunk per file.
        Then, `func_per_chunk` is performed on each sub-view.
        Besides the view, the counter `i` is passed to the `func_per_chunk`,
        which can be used for logging.
        Additional parameters for `func_per_chunk` can be specified using `functools.partial`.
        The computation of each chunk has to be independent of each other.
        Therefore, the work can be parallelized with `executor`.

        If the `View` is of type `MagView` only the bounding box from the properties is chunked.

        Example:
        ```python
        from webknossos.utils import named_partial

        def some_work(args: Tuple[View, int], some_parameter: int) -> None:
            view_of_single_chunk, i = args
            # perform operations on the view
            ...

        # ...
        # let 'mag1' be a `MagView`
        func = named_partial(some_work, some_parameter=42)
        mag1.for_each_chunk(
            func,
        )
        ```
        """

        if chunk_shape is None:
            if chunk_size is not None:
                warn_deprecated("chunk_size", "chunk_shape")
                chunk_shape = Vec3Int(chunk_size)
                self._check_chunk_shape(chunk_shape, read_only=self.read_only)
            else:
                chunk_shape = self._get_file_dimensions_mag1()
        else:
            chunk_shape = Vec3Int(chunk_shape)
            self._check_chunk_shape(chunk_shape, read_only=self.read_only)

        job_args = []
        for i, chunk in enumerate(self.bounding_box.chunk(chunk_shape, chunk_shape)):
            chunk_view = self.get_view(
                absolute_offset=chunk.topleft,
                size=chunk.size,
            )
            job_args.append((chunk_view, i))

        # execute the work for each chunk
        if executor is None:
            if progress_desc is None:
                for args in job_args:
                    func_per_chunk(args)
            else:
                with get_rich_progress() as progress:
                    task = progress.add_task(
                        progress_desc, total=self.bounding_box.volume()
                    )
                    for args in job_args:
                        func_per_chunk(args)
                        current_view: View = args[0]
                        progress.update(
                            task, advance=current_view.bounding_box.volume()
                        )
        else:
            wait_and_ensure_success(
                executor.map_to_futures(func_per_chunk, job_args),
                executor=executor,
                progress_desc=progress_desc,
            )

    def map_chunk(
        self,
        func_per_chunk: Callable[["View"], Any],
        chunk_shape: Optional[Vec3IntLike] = None,  # in Mag(1)
        executor: Optional[Executor] = None,
        progress_desc: Optional[str] = None,
    ) -> List[Any]:
        """
        The view is chunked into multiple sub-views of size `chunk_shape` (in Mag(1)),
        by default one chunk per file.
        Then, `func_per_chunk` is performed on each sub-view and the results are collected
        in a list.
        Additional parameters for `func_per_chunk` can be specified using `functools.partial`.
        The computation of each chunk has to be independent of each other.
        Therefore, the work can be parallelized with `executor`.

        If the `View` is of type `MagView` only the bounding box from the properties is chunked.

        Example:
        ```python
        from webknossos.utils import named_partial

        def some_work(view: View, some_parameter: int) -> None:
            # perform operations on the view
            ...

        # ...
        # let 'mag1' be a `MagView`
        func = named_partial(some_work, some_parameter=42)
        results = mag1.map_chunk(
            func,
        )
        ```
        """

        if chunk_shape is None:
            chunk_shape = self._get_file_dimensions_mag1()
        else:
            chunk_shape = Vec3Int(chunk_shape)
            self._check_chunk_shape(chunk_shape, read_only=self.read_only)

        job_args = []
        for chunk in self.bounding_box.chunk(chunk_shape, chunk_shape):
            chunk_view = self.get_view(
                absolute_offset=chunk.topleft,
                size=chunk.size,
            )
            job_args.append(chunk_view)

        # execute the work for each chunk
        with get_executor_for_args(None, executor) as executor:
            results = wait_and_ensure_success(
                executor.map_to_futures(func_per_chunk, job_args),
                executor=executor,
                progress_desc=progress_desc,
            )

        return results

    def for_zipped_chunks(
        self,
        func_per_chunk: Callable[[Tuple["View", "View", int]], None],
        target_view: "View",
        source_chunk_shape: Optional[Vec3IntLike] = None,  # in Mag(1)
        target_chunk_shape: Optional[Vec3IntLike] = None,  # in Mag(1)
        executor: Optional[Executor] = None,
        progress_desc: Optional[str] = None,
        *,
        source_chunk_size: Optional[Vec3IntLike] = None,  # deprecated
        target_chunk_size: Optional[Vec3IntLike] = None,  # deprecated
    ) -> None:
        """
        This method is similar to `for_each_chunk` in the sense that it delegates work to smaller chunks,
        given by `source_chunk_shape` and `target_chunk_shape` (both in Mag(1),
        by default using the larger of the source_views and the target_views file-sizes).
        However, this method also takes another view as a parameter. Both views are chunked simultaneously
        and a matching pair of chunks is then passed to the function that shall be executed.
        This is useful if data from one view should be (transformed and) written to a different view,
        assuming that the transformation of the data can be handled on chunk-level.
        Additionally to the two views, the counter `i` is passed to the `func_per_chunk`, which can be used for logging.

        The mapping of chunks from the source view to the target is bijective.
        The ratio between the size of the `source_view` (`self`) and the `source_chunk_shape` must be equal to
        the ratio between the `target_view` and the `target_chunk_shape`. This guarantees that the number of chunks
        in the `source_view` is equal to the number of chunks in the `target_view`.
        The `target_chunk_shape` must be a multiple of the file size on disk to avoid concurrent writes.

        Example use case: *downsampling from Mag(1) to Mag(2)*
        - size of the views: `16384³` (`8192³` in Mag(2) for `target_view`)
        - automatic chunk sizes: `2048³`, assuming  default file-lengths
          (`1024³` in Mag(2), which fits the default file-length of 32*32)
        """

        if source_chunk_shape is None and source_chunk_size is not None:
            warn_deprecated("source_chunk_size", "source_chunk_shape")
            source_chunk_shape = source_chunk_size

        if target_chunk_shape is None and target_chunk_size is not None:
            warn_deprecated("target_chunk_size", "target_chunk_shape")
            target_chunk_shape = target_chunk_size

        if source_chunk_shape is None or target_chunk_shape is None:
            assert (
                source_chunk_shape is None and target_chunk_shape is None
            ), "Either both source_chunk_shape and target_chunk_shape must be given or none."
            source_chunk_shape = self._get_file_dimensions_mag1().pairmax(
                target_view._get_file_dimensions_mag1()
            )
            target_chunk_shape = source_chunk_shape
        else:
            source_chunk_shape = Vec3Int(source_chunk_shape)
            target_chunk_shape = Vec3Int(target_chunk_shape)
            self._check_chunk_shape(source_chunk_shape, read_only=True)
            target_view._check_chunk_shape(
                target_chunk_shape, read_only=target_view.read_only
            )

        if self.bounding_box.is_empty() or target_view.bounding_box.is_empty():
            return

        assert np.array_equal(
            self.bounding_box.size.to_np() / target_view.bounding_box.size.to_np(),
            source_chunk_shape.to_np() / target_chunk_shape.to_np(),
        ), (
            "Calling 'for_zipped_chunks' failed because the ratio of the view sizes "
            + f"(source size = {self.bounding_box.size}, target size = {target_view.bounding_box.size}) "
            + "must be equal to the ratio of the chunk sizes "
            + f"(source_chunk_shape in Mag(1) = {source_chunk_shape}, target_chunk_shape in Mag(1) = {target_chunk_shape})"
        )

        job_args = []
        source_chunks = self.bounding_box.chunk(source_chunk_shape, source_chunk_shape)
        target_chunks = target_view.bounding_box.chunk(
            target_chunk_shape, target_chunk_shape
        )

        for i, (source_chunk, target_chunk) in enumerate(
            zip(source_chunks, target_chunks)
        ):
            source_chunk_view = self.get_view(
                absolute_offset=source_chunk.topleft,
                size=source_chunk.size,
                read_only=True,
            )
            target_chunk_view = target_view.get_view(
                absolute_offset=target_chunk.topleft,
                size=target_chunk.size,
            )

            job_args.append((source_chunk_view, target_chunk_view, i))

        # execute the work for each pair of chunks
        if executor is None:
            if progress_desc is None:
                for args in job_args:
                    func_per_chunk(args)
            else:
                with get_rich_progress() as progress:
                    task = progress.add_task(
                        progress_desc, total=self.bounding_box.volume()
                    )
                    for args in job_args:
                        func_per_chunk(args)
                        progress.update(task, advance=args[0].bounding_box.volume())
        else:
            wait_and_ensure_success(
                executor.map_to_futures(func_per_chunk, job_args),
                executor=executor,
                progress_desc=progress_desc,
            )

    def content_is_equal(
        self,
        other: "View",
        args: Optional[Namespace] = None,  # deprecated
        executor: Optional[Executor] = None,
    ) -> bool:
        if args is not None:
            warn_deprecated(
                "args argument",
                "executor (e.g. via webknossos.utils.get_executor_for_args(args))",
            )

        if self.bounding_box.size != other.bounding_box.size:
            return False
        with get_executor_for_args(args, executor) as executor:
            try:
                self.for_zipped_chunks(
                    _assert_check_equality,
                    other,
                    executor=executor,
                    progress_desc="Comparing contents",
                )
            except AssertionError:
                return False
        return True

    def _is_compressed(self) -> bool:
        return self.info.compression_mode

    def get_dtype(self) -> np.dtype:
        """
        Returns the dtype per channel of the data. For example `uint8`.
        """
        return self.info.voxel_type

    def __enter__(self) -> "View":
        warnings.warn(
            "[DEPRECATION] Entering a View to open it is deprecated. The internal dataset will be opened automatically.",
            DeprecationWarning,
        )
        return self

    def __exit__(
        self,
        _type: Optional[Type[BaseException]],
        _value: Optional[BaseException],
        _tb: Optional[TracebackType],
    ) -> None:
        pass

    def __repr__(self) -> str:
        return f"View({repr(str(self._path))}, bounding_box={self.bounding_box})"

    def _check_chunk_shape(self, chunk_shape: Vec3Int, read_only: bool) -> None:
        assert chunk_shape.is_positive(
            strictly_positive=True
        ), f"The passed parameter 'chunk_shape' {chunk_shape} contains at least one 0. This is not allowed."

        divisor = self.mag.to_vec3_int() * self.info.chunk_shape
        if not read_only:
            divisor *= self.info.chunks_per_shard
        assert chunk_shape % divisor == Vec3Int.zeros(), (
            f"The chunk_shape {chunk_shape} must be a multiple of "
            + f"mag*chunk_shape{'*chunks_per_shard' if not read_only else ''} of the view, "
            + f"which is {divisor})."
        )

    def _get_file_dimensions(self) -> Vec3Int:
        return self.info.shard_shape

    def _get_file_dimensions_mag1(self) -> Vec3Int:
        return Vec3Int(self._get_file_dimensions() * self.mag.to_vec3_int())

    @property
    def _array(self) -> BaseArray:
        if self._cached_array is None:
            cls_array = BaseArray.get_class(self.info.data_format)
            self._cached_array = cls_array(self._path)
        return self._cached_array

    @_array.deleter
    def _array(self) -> None:
        if self._cached_array is not None:
            self._cached_array.close()
            self._cached_array = None

    def __del__(self) -> None:
        if hasattr(self, "_cached_array"):
            del self._cached_array

    def __getstate__(self) -> Dict[str, Any]:
        d = dict(self.__dict__)
        del d["_cached_array"]
        return d

    def __setstate__(self, d: Dict[str, Any]) -> None:
        d["_cached_array"] = None
        self.__dict__ = d


def _count_defined_values(values: Iterable[Optional[Any]]) -> int:
    return sum(i is not None for i in values)


def _copy_job(args: Tuple[View, View, int]) -> None:
    (source_view, target_view, _) = args
    # Copy the data form one view to the other in a buffered fashion
    target_view.write(source_view.read())



from typing import Optional

import attr

from ..client.api_client.models import ApiAnnotation
from ..utils import warn_deprecated
from .annotation import Annotation, AnnotationState, AnnotationType


@attr.frozen
class AnnotationInfo:
    """Data class containing information about a WEBKNOSSOS annotation"""

    id: str
    owner_id: Optional[str]
    name: str
    description: str
    type: AnnotationType
    state: AnnotationState
    duration_in_seconds: Optional[float]
    modified: Optional[int]

    def download_annotation(self) -> Annotation:
        """Downloads and returns the annotation that is discribed by this AnnotationInfo object"""
        return Annotation.download(self.id)

    @classmethod
    def _from_api_annotation(cls, api_annotation: ApiAnnotation) -> "AnnotationInfo":
        owner_id = api_annotation.owner.id if api_annotation.owner is not None else None
        return AnnotationInfo(
            id=api_annotation.id,
            owner_id=owner_id,
            name=api_annotation.name,
            description=api_annotation.description,
            type=AnnotationType(api_annotation.typ),
            state=AnnotationState(api_annotation.state),
            duration_in_seconds=(api_annotation.tracing_time / 1000)
            if api_annotation.tracing_time is not None
            else None,
            modified=api_annotation.modified,
        )

    @property
    def user_id(self) -> Optional[str]:
        warn_deprecated("user_id", "owner_id")
        return self.owner_id



"""This module takes care of compressing WEBKNOSSOS datasets."""

from argparse import Namespace
from multiprocessing import cpu_count
from typing import Any, Optional

import typer
from typing_extensions import Annotated

from ..dataset import Dataset
from ..utils import get_executor_for_args
from ._utils import DistributionStrategy, parse_path


def main(
    *,
    target: Annotated[
        Any,
        typer.Argument(
            help="Path to your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Compress a given WEBKNOSSOS dataset."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    with get_executor_for_args(args=executor_args) as executor:
        Dataset.open(target).compress(executor=executor)



"""This module delivers a CLI to work with WEBKNOSSOS datasets."""

import typer

from . import (
    check_equality,
    compress,
    convert,
    convert_knossos,
    convert_raw,
    convert_zarr,
    download,
    downsample,
    export_wkw_as_tiff,
    merge_fallback,
    upload,
    upsample,
)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_short=False)

app.command("check-equality")(check_equality.main)
app.command("compress")(compress.main)
app.command("convert")(convert.main)
app.command("convert-knossos")(convert_knossos.main)
app.command("convert-raw")(convert_raw.main)
app.command("convert-zarr")(convert_zarr.main)
app.command("download")(download.main)
app.command("downsample")(downsample.main)
app.command("export-wkw-as-tiff")(export_wkw_as_tiff.main)
app.command("merge-fallback")(merge_fallback.main)
app.command("upload")(upload.main)
app.command("upsample")(upsample.main)



import warnings
from contextlib import contextmanager, nullcontext
from itertools import chain
from os import PathLike
from pathlib import Path
from typing import (
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)
from urllib.error import HTTPError

import numpy as np
from natsort import natsorted
from numpy.typing import DTypeLike

from webknossos.geometry.bounding_box import BoundingBox
from webknossos.geometry.nd_bounding_box import NDBoundingBox

# pylint: disable=unused-import
try:
    from .pims_czi_reader import PimsCziReader
except ImportError:
    PimsCziReader = type(None)  # type: ignore[misc,assignment]

try:
    from .pims_dm_readers import (  # noqa: F401 unused-import
        PimsDm3Reader,
        PimsDm4Reader,
    )
except ImportError:
    pass

try:
    from .pims_imagej_tiff_reader import (  # noqa: F401 unused-import
        PimsImagejTiffReader,
    )
except ImportError:
    pass


from ...geometry.vec_int import VecInt
from ..mag_view import MagView

try:
    import pims
except ImportError as import_error:
    raise RuntimeError(
        "Cannot import pims, please install it e.g. using 'webknossos[all]'"
    ) from import_error


# Fix ImageIOReader not handling channels correctly. This might get fixed via
# https://github.com/soft-matter/pims/pull/430
pims.ImageIOReader.frame_shape = pims.FramesSequenceND.frame_shape


def _assume_color_channel(dim_size: int, dtype: np.dtype) -> bool:
    return dim_size == 1 or (dim_size == 3 and dtype == np.dtype("uint8"))


class PimsImages:
    dtype: DTypeLike
    num_channels: int

    def __init__(
        self,
        images: Union[str, Path, "pims.FramesSequence", List[Union[str, PathLike]]],
        channel: Optional[int],
        timepoint: Optional[int],
        czi_channel: Optional[int],
        swap_xy: bool,
        flip_x: bool,
        flip_y: bool,
        flip_z: bool,
        use_bioformats: Optional[bool],
        is_segmentation: bool,
    ) -> None:
        """
        During initialization the pims objects are examined and configured to produce
        ndarrays that follow the following form:
        (self._iter_axes, *self._bundle_axis)
        self._iter_axes can be a list of different axes or an empty list if the image is 2D.
        In the latter case, the inner 2D image is still wrapped in a single-element list
        by _open_images() to be consistent with 3D images.
        self._bundle_axis can consist of "x", "y" and "c", where "c" is optional and must be
        at the start or the end, so one of "xy", "yx", "xyc", "yxc", "cxy", "cyx".

        The part "IDENTIFY AXIS ORDER" figures out (self._iter_dim, *self._img_dims)
        from out-of-the-box pims images. Afterwards self._open_images() produces
        images consistent with those variables.

        The part "IDENTIFY SHAPE & CHANNELS" uses this information and the well-defined
        images to figure out the shape & num_channels.
        """
        ## we use images as the name for the entered contextmanager,
        ## the `del` prevents any confusion with the passed argument.
        self._original_images = images
        del images

        ## arguments as inner attributes
        self._channel = channel
        self._timepoint = timepoint
        self._czi_channel = czi_channel
        self._swap_xy = swap_xy
        self._flip_x = flip_x
        self._flip_y = flip_y
        self._flip_z = flip_z
        self._use_bioformats = use_bioformats

        ## attributes that will be set in __init__()
        # _bundle_axes
        self._iter_axes = None
        self._iter_loop_size = None
        self._possible_layers = {}

        ## attributes only for pims.FramesSequenceND instances:
        # _default_coords
        # _init_c_axis

        ## attributes that will also be set in __init__()
        # dtype
        # num_channels
        # _first_n_channels

        #######################
        # IDENTIFY AXIS ORDER #
        #######################

        with self._open_images() as images:
            assert isinstance(
                images, pims.FramesSequence
            ), f"{type(images)} does not inherit from pims.FramesSequence"
            self.dtype = images.dtype

            if isinstance(images, pims.FramesSequenceND):
                self._default_coords = {}
                self._init_c_axis = False
                if isinstance(images, pims.imageio_reader.ImageIOReader):
                    # bugfix for ImageIOReader which misses channel axis sometimes,
                    # assuming channels come last. This might get fixed via
                    # https://github.com/soft-matter/pims/pull/430
                    if (
                        len(images._shape) >= len(images.sizes)
                        and "c" not in images.sizes
                    ):
                        images._init_axis("c", images._shape[-1])
                        self._init_c_axis = True

                if isinstance(images, PimsCziReader):
                    available_czi_channels = images.available_czi_channels()
                    if len(available_czi_channels) > 1:
                        self._possible_layers["czi_channel"] = available_czi_channels

                # An image slice should always consist of a 2D image. If there are multiple channels
                # the data of each channel is part of the image slices. Possible shapes of an image
                # slice are (#y_shape, #x_shape), (1, #y_shape, #x_shape) or (3, #y_shape, #x_shape).
                if images.sizes.get("c", 1) > 1:
                    self._bundle_axes = ["c", "y", "x"]
                else:
                    if "c" in images.axes:
                        # When c-axis is not in _bundle_axes and _iter_axes its value at coordinate 0
                        # should be returned
                        self._default_coords["c"] = 0
                    self._bundle_axes = ["y", "x"]

                # All other axes are used to iterate over them. The last one is iterated the fastest.
                self._iter_axes = list(
                    set(images.axes).difference({*self._bundle_axes, "c", "z"})
                )
                if "z" in images.axes:
                    self._iter_axes.append("z")

                if self._timepoint is not None:
                    # if a timepoint is given, PimsImages should only generate image slices for that timepoint
                    if "t" in self._iter_axes:
                        self._iter_axes.remove("t")
                        self._default_coords["t"] = self._timepoint

                if len(self._iter_axes) > 1:
                    iter_size = 1
                    self._iter_loop_size = dict()
                    for axis, other_axis in zip(
                        self._iter_axes[-1:0:-1], self._iter_axes[-2::-1]
                    ):
                        # Creates a dict that contains the size of the loop for each axis
                        # the axes are identified by their index in the _iter_axes list
                        # the last axis is the fastest iterating axis, therfore the size of the loop
                        # for the last axis is 1. For all other axes it is the product of all previous axes sizes.
                        # self._iter_axes[-1:0:-1] is a reversed copy of self._iter_axes without the last element
                        # e.g. [1,2,3,4] -> [4,3,2]
                        # self._iter_axes[-2::-1] is a reversed copy of self._iter_axes without the first element
                        # e.g. [1,2,3,4] -> [3,2,1]
                        self._iter_loop_size[other_axis] = (
                            iter_size := iter_size * images.sizes[axis]
                        )

            else:
                # Fallback for generic pims classes that do not name their
                # dimensions as pims.FramesSequenceND does:

                _allow_channels_first = not is_segmentation
                if isinstance(images, (pims.ImageSequence, pims.ReaderSequence)):
                    _allow_channels_first = False

                if len(images.shape) == 2:
                    # Assume yx
                    self._bundle_axes = ["y", "x"]
                    self._iter_axes = []
                elif len(images.shape) == 3:
                    # Assume yxc, cyx or zyx
                    if _assume_color_channel(images.shape[2], images.dtype):
                        self._bundle_axes = ["y", "x", "c"]
                        self._iter_axes = []
                    elif images.shape[0] == 1 or (
                        _allow_channels_first
                        and _assume_color_channel(images.shape[0], images.dtype)
                    ):
                        self._bundle_axes = ["c", "y", "x"]
                        self._iter_axes = []
                    else:
                        self._bundle_axes = ["y", "x"]
                        self._iter_axes = ["z"]
                elif len(images.shape) == 4:
                    # Assume zcyx or zyxc
                    if images.shape[1] == 1 or _assume_color_channel(
                        images.shape[1], images.dtype
                    ):
                        self._bundle_axes = ["c", "y", "x"]
                    else:
                        self._bundle_axes = ["y", "x", "c"]
                    self._iter_axes = ["z"]
                elif len(images.shape) == 5:
                    # Assume tzcyx or tzyxc
                    # t has to be constant for this reader to obtain 4D image
                    # (only possible if not specified manually already, since
                    # the timepoint would already be indexed here and the
                    # 5th dimension would be something else)
                    if timepoint is not None:
                        raise RuntimeError(
                            f"Got {len(images.shape)} axes for the images after "
                            + "removing time dimension, can only map to 3D+channels."
                            + "To import image with more dimensions use dataformat"
                            + "Zarr3 and set use_bioformats=True."
                        )

                    if _assume_color_channel(images.shape[2], images.dtype):
                        self._bundle_axes = ["c", "y", "x"]
                    else:
                        self._bundle_axes = ["y", "x", "c"]
                    self._iter_axes = ["z"]
                    self._timepoint = 0
                    if images.shape[0] > 1:
                        self._possible_layers["timepoint"] = list(
                            range(0, images.shape[0])
                        )
                else:
                    raise RuntimeError(
                        f"Got {len(images.shape)} axes for the images, "
                        + "but don't have axes information. Try to open "
                        + "an N-dimensional image file with use_bioformats="
                        + "True."
                    )

        #########################
        # IDENTIFY NUM_CHANNELS #
        #########################

        with self._open_images() as images:
            try:
                c_index = self._bundle_axes.index("c")
                if isinstance(images, list):
                    images_shape = (len(images),) + cast(
                        pims.FramesSequence, images[0]
                    ).shape
                else:
                    images_shape = images.shape  # pylint: disable=no-member

                self.num_channels = images_shape[c_index + 1]

            except ValueError:
                self.num_channels = 1

        self._first_n_channels = None
        if self._channel is not None:
            assert (
                self._channel < self.num_channels
            ), f"Selected channel {self._channel} (0-indexed), but only {self.num_channels} channels are available."
            self.num_channels = 1
        else:
            if self.num_channels == 2:
                self._possible_layers["channel"] = [0, 1]
                self.num_channels = 1
                self._channel = 0
            elif self.num_channels > 3:
                self._possible_layers["channel"] = list(range(0, self.num_channels))
                self.num_channels = 3
                self._first_n_channels = 3

    def _normalize_original_images(self) -> Union[str, List[str]]:
        original_images = self._original_images
        if isinstance(original_images, (str, Path)):
            original_images_path = Path(original_images)
            if original_images_path.is_dir():
                valid_suffixes = get_valid_pims_suffixes()
                if self._use_bioformats is not False:
                    valid_suffixes.update(get_valid_bioformats_suffixes())
                original_images = natsorted(
                    str(i)
                    for i in original_images_path.glob("**/*")
                    if i.is_file() and i.suffix.lstrip(".") in valid_suffixes
                )
                if len(original_images) == 1:
                    original_images = original_images[0]
        if isinstance(original_images, str):
            return original_images
        elif isinstance(original_images, Iterable):
            return [str(i) for i in original_images]
        else:
            return str(original_images)

    def _ensure_correct_bioformats_usage(
        self, images_context_manager: pims.FramesSequence
    ) -> None:
        if (
            isinstance(images_context_manager, pims.bioformats.BioformatsReader)
            and self._use_bioformats is False
        ):  # None is allowed
            raise RuntimeError(
                "Selected bioformats reader, but using bioformats is not allowed "
                + "(use_bioformats is False)."
            )

    def _try_open_pims_images(
        self, original_images: Union[str, List[str]], exceptions: List[Exception]
    ) -> Optional[pims.FramesSequence]:
        if self._use_bioformats:
            return None

        open_kwargs = {}
        if self._czi_channel is not None:
            open_kwargs["czi_channel"] = self._czi_channel

        # try normal pims.open
        def strategy_0() -> pims.FramesSequence:
            result = pims.open(original_images, **open_kwargs)
            self._ensure_correct_bioformats_usage(original_images)
            return result

        # try pims.ImageSequence, which uses skimage internally but works for multiple images
        strategy_1 = lambda: pims.ImageSequence(original_images)  # noqa: E731 Do not assign a `lambda` expression, use a `def`

        # for image lists, try to guess the correct reader using only the first image,
        # and apply that for all images via pims.ReaderSequence
        def strategy_2() -> Optional[pims.FramesSequence]:
            if isinstance(original_images, list):
                # assuming the same reader works for all images:
                first_image_handler = pims.open(original_images[0], **open_kwargs)
                self._ensure_correct_bioformats_usage(first_image_handler)
                return pims.ReaderSequence(
                    original_images, type(first_image_handler), **open_kwargs
                )
            else:
                return None

        for strategy in [strategy_0, strategy_1, strategy_2]:
            try:
                images_context_manager = strategy()
            except Exception as e:  # noqa: PERF203 `try`-`except` within a loop incurs performance overhead
                exceptions.append(e)
            else:
                if images_context_manager is not None:
                    return images_context_manager
        return None

    def _try_open_bioformats_images_raw(
        self,
        original_images: Union[str, List[str]],
        exceptions: List[Exception],
    ) -> Optional[pims.FramesSequence]:
        try:
            if self._use_bioformats is False:  # None is allowed
                raise RuntimeError(
                    "Using bioformats is not allowed (use_bioformats is False)."
                )

            # There is a wrong warning about jpype, supressing it here.
            # See issue https://github.com/soft-matter/pims/issues/384
            warnings.filterwarnings(
                "ignore",
                "Due to an issue with JPype 0.6.0, reading is slower.*",
                category=UserWarning,
                module="pims.bioformats",
            )
            try:
                pims.bioformats._find_jar()
            except HTTPError:
                # We cannot use the newest bioformats version,
                # since it does not include the necessary loci_tools.jar.
                # Updates to support newer bioformats jars with pims are in PR
                # https://github.com/soft-matter/pims/pull/403

                # This is also part of the worker dockerfile to cache the
                # jar in the image, please update Dockerfile.worker in the
                # voxelytics repo accordingly when editing this.
                pims.bioformats.download_jar(version="6.7.0")

            if "*" in str(original_images) or isinstance(original_images, list):
                return pims.ReaderSequence(
                    original_images, pims.bioformats.BioformatsReader
                )
            else:
                return pims.bioformats.BioformatsReader(original_images)
        except Exception as e:
            exceptions.append(e)
            return None

    @contextmanager
    def _open_images(
        self,
    ) -> Iterator[Union[pims.FramesSequence, List[pims.FramesSequence]]]:
        """
        This yields well-defined images of the form (self._iter_axes, *self._bundle_axes),
        after IDENTIFY AXIS ORDER of __init__() has run.
        For a 2D image this is achieved by wrapping it in a list.
        """
        images_context_manager: Optional[ContextManager]
        with warnings.catch_warnings():
            if isinstance(self._original_images, pims.FramesSequenceND):
                images_context_manager = nullcontext(enter_result=self._original_images)
            else:
                exceptions: List[Exception] = []
                original_images = self._normalize_original_images()
                images_context_manager = None

                images_context_manager = self._try_open_pims_images(
                    original_images, exceptions
                )

                if images_context_manager is None:
                    images_context_manager = self._try_open_bioformats_images_raw(
                        original_images, exceptions
                    )

                if images_context_manager is None:
                    if len(exceptions) == 1:
                        raise exceptions[0]
                    else:
                        exceptions_str = "\n".join(
                            f"{type(e).__name__}: {str(e)}" for e in exceptions
                        )
                        raise ValueError(
                            f"Tried to open the images {self._original_images} with different methods, "
                            + f"none succeded. The following errors were raised:\n{exceptions_str}"
                        )

            with images_context_manager as images:
                if isinstance(images, pims.FramesSequenceND):
                    if hasattr(self, "_bundle_axes"):
                        # first part of __init__() has happened
                        images.default_coords.update(self._default_coords)
                        if self._init_c_axis and "c" not in images.sizes:
                            # Bugfix for ImageIOReader which misses channel axis sometimes,
                            # assuming channels come last. _init_c_axis is set in __init__().
                            # This might get fixed via https://github.com/soft-matter/pims/pull/430
                            images._init_axis("c", images._shape[-1])
                            for key in list(images._get_frame_dict.keys()):
                                images._get_frame_dict[key + ("c",)] = (
                                    images._get_frame_dict.pop(key)
                                )
                            self._bundle_axes.remove("c")
                            self._bundle_axes.append("c")
                        images.bundle_axes = self._bundle_axes
                        images.iter_axes = self._iter_axes
                else:
                    if self._timepoint is not None:
                        images = images[self._timepoint]
                        if self._iter_axes and "t" in self._iter_axes:
                            self._iter_axes.remove("t")
                    if self._iter_axes == []:
                        # add outer list to wrap 2D images as 3D-like structure
                        images = [images]
                yield images

    def copy_to_view(
        self,
        args: Union[BoundingBox, NDBoundingBox],
        mag_view: MagView,
        is_segmentation: bool,
        dtype: Optional[DTypeLike] = None,
    ) -> Tuple[Tuple[int, int], Optional[int]]:
        """Copies the images according to the passed arguments to the given mag_view.
        args is expected to be a (ND)BoundingBox the start and end of the z-range, meant for usage with an executor.
        copy_to_view returns an iterable of image shapes and largest segment ids. When using this
        method a manual update of the bounding box and the largest segment id might be necessary.
        """
        absolute_bbox = args
        relative_bbox = absolute_bbox.offset(-mag_view.bounding_box.topleft)

        assert all(
            size == 1
            for size, axis in zip(absolute_bbox.size, absolute_bbox.axes)
            if axis not in ("x", "y", "z")
        ), "The delivered BoundingBox has to be flat except for x,y and z dimension."

        # z_start and z_end are relative to the bounding box of the mag_view
        # to access the correct data from the images
        z_start, z_end = relative_bbox.get_bounds("z")
        shapes = []
        max_id: Optional[int]
        if is_segmentation:
            max_id = 0
        else:
            max_id = None

        with self._open_images() as images:
            if self._iter_axes is not None and self._iter_loop_size is not None:
                # select the range of images that represents one xyz combination in the mag_view
                lower_bounds = sum(
                    self._iter_loop_size[axis_name]
                    * relative_bbox.get_bounds(axis_name)[0]
                    for axis_name in self._iter_axes[:-1]
                )
                upper_bounds = lower_bounds + mag_view.bounding_box.get_shape("z")
                images = images[lower_bounds:upper_bounds]
            if self._flip_z:
                images = images[::-1]  # pylint: disable=unsubscriptable-object

            with mag_view.get_buffered_slice_writer(
                # Previously only z_start and its end were important, now the slice writer needs to know
                # which axis is currently written.
                absolute_bounding_box=absolute_bbox,
                buffer_size=absolute_bbox.get_shape("z"),
                # copy_to_view is typically used in a multiprocessing-context. Therefore the
                # buffered slice writer should not update the json file to avoid race conditions.
                json_update_allowed=False,
            ) as writer:
                for image_slice in images[z_start:z_end]:
                    image_slice = np.array(image_slice)
                    # place channels first
                    if "c" in self._bundle_axes:
                        if hasattr(self, "_init_c_axis") and self._init_c_axis:
                            # Bugfix for ImageIOReader which misses channel axis sometimes,
                            # assuming channels come last. _init_c_axis is set in __init__().
                            # This might get fixed via
                            image_slice = image_slice[0]
                        image_slice = np.moveaxis(
                            image_slice,
                            source=self._bundle_axes.index("c"),
                            destination=0,
                        )
                        if self._channel is not None:
                            image_slice = image_slice[self._channel : self._channel + 1]
                        elif self._first_n_channels is not None:
                            image_slice = image_slice[: self._first_n_channels]
                        assert image_slice.shape[0] == self.num_channels, (
                            f"Image shape {image_slice.shape} does not fit to the number of channels "
                            + f"{self.num_channels} which are expected in the first axis."
                        )

                    if self._flip_x:
                        image_slice = np.flip(image_slice, -2)
                    if self._flip_y:
                        image_slice = np.flip(image_slice, -1)

                    if dtype is not None:
                        image_slice = image_slice.astype(dtype, order="F")

                    if max_id is not None:
                        max_id = max(max_id, image_slice.max())

                    if self._swap_xy is False:
                        image_slice = np.moveaxis(image_slice, -1, -2)

                    shapes.append(image_slice.shape[-2:])
                    writer.send(image_slice)

            return dimwise_max(shapes), None if max_id is None else int(max_id)

    def get_possible_layers(self) -> Optional[Dict["str", List[int]]]:
        if len(self._possible_layers) == 0:
            return None
        else:
            return self._possible_layers

    @property
    def expected_bbox(self) -> NDBoundingBox:
        # replaces the previous expected_shape to enable n-dimensional input files
        with self._open_images() as images:
            if isinstance(images, pims.FramesSequenceND):
                axes = images.axes
                images_shape = tuple(images.sizes[axis] for axis in axes)
            else:
                if isinstance(images, list):
                    images_shape = (len(images),) + cast(
                        pims.FramesSequence, images[0]
                    ).shape

                else:
                    images_shape = images.shape  # pylint: disable=no-member
                if len(images_shape) == 3:
                    axes = ("z", "y", "x")
                else:
                    axes = ("z", "c", "y", "x")

            if self._iter_loop_size is None:
                # There is no or only one element in self._iter_axes, so a 3D bounding box is sufficient.
                x_index, y_index = (
                    axes.index("x"),
                    axes.index("y"),
                )
                if self._iter_axes:
                    try:
                        # In case the naming of the third axis is not "z",
                        # it is still considered as the z-axis.
                        z_index = axes.index(self._iter_axes[0])
                    except ValueError:
                        z_index = axes.index("z")
                    z_shape = images_shape[z_index]
                else:
                    z_shape = 1
                if self._swap_xy:
                    x_index, y_index = y_index, x_index
                return BoundingBox(
                    (0, 0, 0),
                    (images_shape[x_index], images_shape[y_index], z_shape),
                )
            else:
                if isinstance(images, pims.FramesSequenceND):
                    axes_names = (self._iter_axes or []) + [
                        axis for axis in self._bundle_axes if axis != "c"
                    ]
                    axes_sizes = [
                        images.sizes[axis]  # pylint: disable=no-member
                        for axis in axes_names
                    ]
                    axes_index = list(range(1, len(axes_names) + 1))
                    topleft = VecInt.zeros(tuple(axes_names))

                    if self._swap_xy:
                        x_index, y_index = axes_names.index("x"), axes_names.index("y")
                        axes_sizes[x_index], axes_sizes[y_index] = (
                            axes_sizes[y_index],
                            axes_sizes[x_index],
                        )

                    return NDBoundingBox(
                        topleft,
                        VecInt(axes_sizes, axes=axes_names),
                        axes_names,
                        VecInt(axes_index, axes=axes_names),
                    )

                raise ValueError(
                    "It seems as if you try to load an N-dimensional image from 2D images. This is currently not supported."
                )


T = TypeVar("T", bound=Tuple[int, ...])


def dimwise_max(vectors: Sequence[T]) -> T:
    if len(vectors) == 1:
        return vectors[0]
    else:
        return cast(T, tuple(map(max, *vectors)))


C = TypeVar("C", bound=Type)


def _recursive_subclasses(cls: C) -> List[C]:
    "Return all subclasses (and their subclasses, etc.)."
    # Source: http://stackoverflow.com/a/3862957/1221924
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in _recursive_subclasses(s)
    ]


def _get_all_pims_handlers() -> (
    Iterable[Type[Union[pims.FramesSequence, pims.FramesSequenceND]]]
):
    return chain(
        _recursive_subclasses(pims.FramesSequence),
        _recursive_subclasses(pims.FramesSequenceND),
    )


def get_valid_pims_suffixes() -> Set[str]:
    valid_suffixes = set()
    for pims_handler in _get_all_pims_handlers():
        valid_suffixes.update(pims_handler.class_exts())
    return valid_suffixes


def get_valid_bioformats_suffixes() -> Set[str]:
    # Added the most present suffixes that are implemented in bioformats
    return {
        "dcm",
        "dicom",
        "ics",
        "ids",
        "lei",
        "tif",
        "lif",
        "stk",
        "nd",
        "nd2",
        "png",
        "tiff",
        "tf2",
        "tf8",
        "btf",
        "pic",
        "raw",
        "xml",
        "gif",
        "nii",
    }


def has_image_z_dimension(
    filepath: Path,
    use_bioformats: Optional[bool],
    is_segmentation: bool,
) -> bool:
    pims_images = PimsImages(
        filepath,
        use_bioformats=use_bioformats,
        is_segmentation=is_segmentation,
        # the following arguments shouldn't matter much for the Dataset.from_images method:
        channel=None,
        timepoint=None,
        czi_channel=None,
        swap_xy=False,
        flip_x=False,
        flip_y=False,
        flip_z=False,
    )

    return pims_images.expected_bbox.get_shape("z") > 1



"""This module converts an image stack to a WEBKNOSSOS dataset."""

from argparse import Namespace
from multiprocessing import cpu_count
from typing import Any, Optional

import typer
from typing_extensions import Annotated

from ..dataset import DataFormat, Dataset
from ..utils import get_executor_for_args
from ._utils import (
    DistributionStrategy,
    LayerCategory,
    VoxelSize,
    parse_path,
    parse_voxel_size,
)


def main(
    *,
    source: Annotated[
        Any,
        typer.Argument(
            help="Path to your image data.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    target: Annotated[
        Any,
        typer.Argument(
            help="Target path to save your WEBKNOSSOS dataset.",
            show_default=False,
            parser=parse_path,
        ),
    ],
    voxel_size: Annotated[
        VoxelSize,
        typer.Option(
            help="The size of one voxel in source data in nanometers. "
            "Should be a comma seperated string (e.g. 11.0,11.0,20.0).",
            parser=parse_voxel_size,
            metavar="VOXEL_SIZE",
        ),
    ],
    layer_name: Annotated[
        Optional[str],
        typer.Option(
            help="This name is used if only one layer is created. Otherwise this name is used as a common prefix for all layers.",
        ),
    ] = None,
    category: Annotated[
        Optional[LayerCategory],
        typer.Option(
            help="The category of the layer that should be created.",
        ),
    ] = None,
    data_format: Annotated[
        DataFormat,
        typer.Option(
            help="Data format to store the target dataset in.",
        ),
    ] = "wkw",  # type:ignore
    name: Annotated[
        Optional[str],
        typer.Option(
            help="New name for the WEBKNOSSOS dataset "
            "(if not provided, final component of target path is used)"
        ),
    ] = None,
    compress: Annotated[
        bool, typer.Option(help="Enable compression of the target dataset.")
    ] = False,
    jobs: Annotated[
        int,
        typer.Option(
            help="Number of processes to be spawned.",
            rich_help_panel="Executor options",
        ),
    ] = cpu_count(),
    distribution_strategy: Annotated[
        DistributionStrategy,
        typer.Option(
            help="Strategy to distribute the task across CPUs or nodes.",
            rich_help_panel="Executor options",
        ),
    ] = DistributionStrategy.MULTIPROCESSING,
    job_resources: Annotated[
        Optional[str],
        typer.Option(
            help="Necessary when using slurm as distribution strategy. Should be a JSON string "
            '(e.g., --job_resources=\'{"mem": "10M"}\')\'',
            rich_help_panel="Executor options",
        ),
    ] = None,
) -> None:
    """Automatic detection of an image stack and conversion to a WEBKNOSSOS dataset."""

    executor_args = Namespace(
        jobs=jobs,
        distribution_strategy=distribution_strategy.value,
        job_resources=job_resources,
    )

    with get_executor_for_args(args=executor_args) as executor:
        Dataset.from_images(
            source,
            target,
            voxel_size,
            name,
            data_format=data_format,
            executor=executor,
            compress=compress,
            layer_name=layer_name,
            layer_category=category.value if category else None,
        )
